# OctoPrint API compatibility
#
# Copyright (C) 2021 Nickolas Grigoriadis <nagrigoriadis@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import logging
from ..common import RequestType, TransportType, KlippyState

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
)
if TYPE_CHECKING:
    from .klippy_connection import KlippyConnection
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .klippy_apis import KlippyAPI as APIComp
    from .file_manager.file_manager import FileManager
    from .job_queue import JobQueue

OCTO_VERSION = '1.5.0'


class OctoPrintCompat:
    """
    Minimal implementation of the REST API as described here:
    https://docs.octoprint.org/en/master/api/index.html

    So that Cura OctoPrint plugin will function for:
    * Handshake
    * Upload gcode/ufp
    * Webcam config
    * Manual GCode submission
    * Heater temperatures
    """

    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.software_version = self.server.get_app_args().get(
            'software_version')
        self.enable_ufp: bool = config.getboolean('enable_ufp', True)

        # Get webcam settings from config
        self.webcam: Dict[str, Any] = {
            'flipH': config.getboolean('flip_h', False),
            'flipV': config.getboolean('flip_v', False),
            'rotate90': config.getboolean('rotate_90', False),
            'streamUrl': config.get('stream_url', '/webcam/?action=stream'),
            'webcamEnabled': config.getboolean('webcam_enabled', True),
        }

        # Local variables
        self.klippy_apis: APIComp = self.server.lookup_component('klippy_apis')
        self.heaters: Dict[str, Dict[str, Any]] = {}
        self.last_print_stats: Dict[str, Any] = {}

        # Register status update event
        self.server.register_event_handler(
            'server:klippy_ready', self._init)
        self.server.register_event_handler(
            'server:status_update', self._handle_status_update)

        # Version & Server information
        self.server.register_endpoint(
            '/api/version', RequestType.GET, self._get_version,
            transports=TransportType.HTTP, wrap_result=False
        )
        self.server.register_endpoint(
            '/api/server', RequestType.GET, self._get_server,
            transports=TransportType.HTTP, wrap_result=False
        )

        # Login, User & Settings
        self.server.register_endpoint(
            '/api/login', RequestType.POST, self._post_login_user,
            transports=TransportType.HTTP, wrap_result=False
        )
        self.server.register_endpoint(
            '/api/currentuser', RequestType.GET, self._post_login_user,
            transports=TransportType.HTTP, wrap_result=False
        )
        self.server.register_endpoint(
            '/api/settings', RequestType.GET, self._get_settings,
            transports=TransportType.HTTP, wrap_result=False
        )

        # File operations
        # Note that file upload is handled in file_manager.py
        # TODO: List/info/select/delete files

        # Job operations
        self.server.register_endpoint(
            '/api/job', RequestType.GET, self._get_job,
            transports=TransportType.HTTP, wrap_result=False
        )
        # TODO: start/cancel/restart/pause jobs

        # Printer operations
        self.server.register_endpoint(
            '/api/printer', RequestType.GET, self._get_printer,
            transports=TransportType.HTTP, wrap_result=False)
        self.server.register_endpoint(
            '/api/printer/command', RequestType.POST, self._post_command,
            transports=TransportType.HTTP, wrap_result=False
        )
        # TODO: head/tool/bed/chamber specific read/issue

        # Printer profiles
        self.server.register_endpoint(
            '/api/printerprofiles', RequestType.GET, self._get_printerprofiles,
            transports=TransportType.HTTP, wrap_result=False
        )

        # Upload Handlers
        self.server.register_upload_handler(
            "/api/files/local", location_prefix="api/files/moonraker")
        self.server.register_endpoint(
            "/api/files/moonraker/(?P<relative_path>.+)", RequestType.POST,
            self._select_file, transports=TransportType.HTTP, wrap_result=False
        )

        # System
        # TODO: shutdown/reboot/restart operations

    async def _init(self) -> None:
        self.heaters = {}
        # Fetch heaters
        try:
            result: Dict[str, Any]
            sensors: List[str]
            result = await self.klippy_apis.query_objects({'heaters': None})
            sensors = result.get('heaters', {}).get('available_sensors', [])
        except self.server.error as e:
            logging.info(f'Error Configuring heaters: {e}')
            sensors = []
        # subscribe objects
        sub: Dict[str, Any] = {s: None for s in sensors}
        sub['print_stats'] = None
        result = await self.klippy_apis.subscribe_objects(sub)
        self.last_print_stats = result.get('print_stats', {})
        if sensors:
            self.heaters = {name: result.get(name, {}) for name in sensors}

    def _handle_status_update(self, status: Dict[str, Any]) -> None:
        if 'print_stats' in status:
            self.last_print_stats.update(status['print_stats'])
        for heater_name, data in self.heaters.items():
            if heater_name in status:
                data.update(status[heater_name])

    def printer_state(self) -> str:
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        klippy_state = kconn.state
        if not klippy_state.startup_complete():
            return 'Offline'
        elif klippy_state != KlippyState.READY:
            return 'Error'
        return {
            'standby': 'Operational',
            'printing': 'Printing',
            'paused': 'Paused',
            'complete': 'Operational'
        }.get(self.last_print_stats.get('state', 'standby'), 'Error')

    def printer_temps(self) -> Dict[str, Any]:
        temps: Dict[str, Any] = {}
        for heater, data in self.heaters.items():
            name = 'bed'
            if heater.startswith('extruder'):
                try:
                    tool_no = int(heater[8:])
                except ValueError:
                    tool_no = 0
                name = f'tool{tool_no}'
            elif heater != "heater_bed":
                continue
            temps[name] = {
                'actual': round(data.get('temperature', 0.), 2),
                'offset': 0,
                'target': data.get('target', 0.),
            }
        return temps

    async def _get_version(self,
                           web_request: WebRequest
                           ) -> Dict[str, str]:
        """
        Version information
        """
        return {
            'server': OCTO_VERSION,
            'api': '0.1',
            'text': f'OctoPrint (Moonraker {self.software_version})',
        }

    async def _get_server(self,
                          web_request: WebRequest
                          ) -> Dict[str, Any]:
        """
        Server status
        """
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        klippy_state = kconn.state
        return {
            'server': OCTO_VERSION,
            'safemode': None if klippy_state == KlippyState.READY else 'settings'
        }

    async def _post_login_user(self,
                               web_request: WebRequest
                               ) -> Dict[str, Any]:
        """
        Confirm session login.

        Since we only support apikey auth, do nothing.
        Report hardcoded user called _api
        """
        return {
            '_is_external_client': False,
            '_login_mechanism': 'apikey',
            'name': '_api',
            'active': True,
            'user': True,
            'admin': True,
            'apikey': None,
            'permissions': [],
            'groups': ['admins', 'users'],
        }

    async def _get_settings(self,
                            web_request: WebRequest
                            ) -> Dict[str, Any]:
        """
        Used to parse OctoPrint capabilities
        """
        settings = {
            'plugins': {},
            'feature': {
                'sdSupport': False,
                'temperatureGraph': False
            },
            'webcam': self.webcam,
        }
        if self.enable_ufp:
            settings['plugins'] = {
                'UltimakerFormatPackage': {
                    'align_inline_thumbnail': False,
                    'inline_thumbnail': False,
                    'inline_thumbnail_align_value': 'left',
                    'inline_thumbnail_scale_value': '50',
                    'installed': True,
                    'installed_version': '0.2.2',
                    'scale_inline_thumbnail': False,
                    'state_panel_thumbnail': True,
                },
            }
        return settings

    async def _get_job(self,
                       web_request: WebRequest
                       ) -> Dict[str, Any]:
        """
        Get current job status
        """
        return {
            'job': {
                'file': {'name': None},
                'estimatedPrintTime': None,
                'filament': {'length': None},
                'user': None,
            },
            'progress': {
                'completion': None,
                'filepos': None,
                'printTime': None,
                'printTimeLeft': None,
                'printTimeOrigin': None,
            },
            'state': self.printer_state()
        }

    async def _get_printer(self,
                           web_request: WebRequest
                           ) -> Dict[str, Any]:
        """
        Get Printer status
        """
        state = self.printer_state()
        return {
            'temperature': self.printer_temps(),
            'state': {
                'text': state,
                'flags': {
                    'operational': state not in ['Error', 'Offline'],
                    'paused': state == 'Paused',
                    'printing': state == 'Printing',
                    'cancelling': state == 'Cancelling',
                    'pausing': False,
                    'error': state == 'Error',
                    'ready': state == 'Operational',
                    'closedOrError': state in ['Error', 'Offline'],
                },
            },
        }

    async def _post_command(self,
                            web_request: WebRequest
                            ) -> Dict:
        """
        Request to run some gcode command
        """
        commands: List[str] = web_request.get('commands', [])
        for command in commands:
            logging.info(f'Executing GCode: {command}')
            try:
                await self.klippy_apis.run_gcode(command)
            except self.server.error:
                msg = f"Error executing GCode {command}"
                logging.exception(msg)

        return {}

    async def _get_printerprofiles(self,
                                   web_request: WebRequest
                                   ) -> Dict[str, Any]:
        """
        Get Printer profiles
        """
        return {
            'profiles': {
                '_default': {
                    'id': '_default',
                    'name': 'Default',
                    'color': 'default',
                    'model': 'Default',
                    'default': True,
                    'current': True,
                    'heatedBed': 'heater_bed' in self.heaters,
                    'heatedChamber': 'chamber' in self.heaters,
                    'axes': {
                        'x': {
                            'speed': 6000.,
                            'inverted': False
                        },
                        'y': {
                            'speed': 6000.,
                            'inverted': False
                        },
                        'z': {
                            'speed': 6000.,
                            'inverted': False
                        },
                        'e': {
                            'speed': 300.,
                            'inverted': False
                        }
                    }
                }
            }
        }

    async def _select_file(self,
                           web_request: WebRequest
                           ) -> None:
        command: str = web_request.get_str('command')
        rel_path: str = web_request.get_str('relative_path')
        root, filename = rel_path.strip("/").split("/", 1)
        fmgr: FileManager = self.server.lookup_component('file_manager')
        if command == "select":
            start_print: bool = web_request.get_boolean('print', False)
            if not start_print:
                # No-op, selecting a file has no meaning in Moonraker
                return
            if root != "gcodes":
                raise self.server.error(
                    "File must be located in the 'gcodes' root", 400)
            if not fmgr.check_file_exists(root, filename):
                raise self.server.error("File does not exist")
            try:
                ret = await self.klippy_apis.query_objects(
                    {'print_stats': None})
                pstate: str = ret['print_stats']['state']
            except self.server.error:
                pstate = "not_avail"
            started: bool = False
            user = web_request.get_current_user()
            if pstate not in ["printing", "paused", "not_avail"]:
                try:
                    await self.klippy_apis.start_print(filename, user=user)
                except self.server.error:
                    started = False
                else:
                    logging.debug(f"Job '{filename}' started via OctoPrint API")
                    started = True
            if not started:
                if fmgr.upload_queue_enabled():
                    job_queue: JobQueue = self.server.lookup_component(
                        'job_queue')
                    await job_queue.queue_job(filename, check_exists=False, user=user)
                    logging.debug(f"Job '{filename}' queued via OctoPrint API")
                else:
                    raise self.server.error("Conflict", 409)
        else:
            raise self.server.error(f"Unsupported Command: {command}")


def load_component(config: ConfigHelper) -> OctoPrintCompat:
    return OctoPrintCompat(config)



# Klipper data logging and storage storage
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import logging
import time
from collections import deque
from ..common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
    Deque,
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .klippy_connection import KlippyConnection
    from .klippy_apis import KlippyAPI as APIComp
    GCQueue = Deque[Dict[str, Any]]
    TempStore = Dict[str, Dict[str, Deque[Optional[float]]]]

TEMP_UPDATE_TIME = 1.

def _round_null(val: Optional[float], ndigits: int) -> Optional[float]:
    if val is None:
        return val
    return round(val, ndigits)

class DataStore:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.temp_store_size = config.getint('temperature_store_size', 1200)
        self.gcode_store_size = config.getint('gcode_store_size', 1000)

        # Temperature Store Tracking
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        self.subscription_cache = kconn.get_subscription_cache()
        self.gcode_queue: GCQueue = deque(maxlen=self.gcode_store_size)
        self.temperature_store: TempStore = {}
        self.temp_monitors: List[str] = []
        eventloop = self.server.get_event_loop()
        self.temp_update_timer = eventloop.register_timer(
            self._update_temperature_store)

        self.server.register_event_handler(
            "server:gcode_response", self._update_gcode_store)
        self.server.register_event_handler(
            "server:klippy_ready", self._init_sensors)
        self.server.register_event_handler(
            "klippy_connection:gcode_received", self._store_gcode_command
        )

        # Register endpoints
        self.server.register_endpoint(
            "/server/temperature_store", RequestType.GET,
            self._handle_temp_store_request
        )
        self.server.register_endpoint(
            "/server/gcode_store", RequestType.GET,
            self._handle_gcode_store_request
        )

    async def _init_sensors(self) -> None:
        klippy_apis: APIComp = self.server.lookup_component('klippy_apis')
        # Fetch sensors
        try:
            result: Dict[str, Any]
            result = await klippy_apis.query_objects({'heaters': None})
        except self.server.error as e:
            logging.info(f"Error Configuring Sensors: {e}")
            return
        heaters: Dict[str, List[str]] = result.get("heaters", {})
        sensors = heaters.get("available_sensors", [])
        self.temp_monitors = heaters.get("available_monitors", [])
        sensors.extend(self.temp_monitors)

        if sensors:
            # Add Subscription
            sub: Dict[str, Optional[List[str]]] = {s: None for s in sensors}
            try:
                status: Dict[str, Any]
                status = await klippy_apis.subscribe_objects(sub)
            except self.server.error as e:
                logging.info(f"Error subscribing to sensors: {e}")
                return
            logging.info(f"Configuring available sensors: {sensors}")
            new_store: TempStore = {}
            valid_fields = ("temperature", "target", "power", "speed")
            for sensor in sensors:
                reported_fields = [
                    f for f in list(status.get(sensor, {}).keys()) if f in valid_fields
                ]
                if not reported_fields:
                    logging.info(f"No valid fields reported for sensor: {sensor}")
                    self.temperature_store.pop(sensor, None)
                    continue
                if sensor in self.temperature_store:
                    new_store[sensor] = self.temperature_store[sensor]
                    for field in list(new_store[sensor].keys()):
                        if field not in reported_fields:
                            new_store[sensor].pop(field, None)
                        else:
                            initial_val: Optional[float]
                            initial_val = _round_null(status[sensor][field], 2)
                            new_store[sensor][field].append(initial_val)
                else:
                    new_store[sensor] = {}
                for field in reported_fields:
                    if field not in new_store[sensor]:
                        initial_val = _round_null(status[sensor][field], 2)
                        new_store[sensor][field] = deque(
                            [initial_val], maxlen=self.temp_store_size
                        )
            self.temperature_store = new_store
            self.temp_update_timer.start(delay=1.)
        else:
            logging.info("No sensors found")
            self.temperature_store = {}
            self.temp_monitors = []
            self.temp_update_timer.stop()

    def _update_temperature_store(self, eventtime: float) -> float:
        for sensor_name, sensor in self.temperature_store.items():
            sdata: Dict[str, Any] = self.subscription_cache.get(sensor_name, {})
            for field, store in sensor.items():
                store.append(_round_null(sdata.get(field, store[-1]), 2))
        return eventtime + TEMP_UPDATE_TIME

    async def _handle_temp_store_request(
        self, web_request: WebRequest
    ) -> Dict[str, Dict[str, List[Optional[float]]]]:
        include_monitors = web_request.get_boolean("include_monitors", False)
        store = {}
        for name, sensor in self.temperature_store.items():
            if not include_monitors and name in self.temp_monitors:
                continue
            store[name] = {f"{k}s": list(v) for k, v in sensor.items()}
        return store

    async def close(self) -> None:
        self.temp_update_timer.stop()

    def _update_gcode_store(self, response: str) -> None:
        curtime = time.time()
        self.gcode_queue.append(
            {'message': response, 'time': curtime, 'type': "response"})

    def _store_gcode_command(self, script: str) -> None:
        curtime = time.time()
        for cmd in script.split('\n'):
            cmd = cmd.strip()
            if not cmd:
                continue
            self.gcode_queue.append(
                {'message': script, 'time': curtime, 'type': "command"})

    async def _handle_gcode_store_request(self,
                                          web_request: WebRequest
                                          ) -> Dict[str, List[Dict[str, Any]]]:
        count = web_request.get_int("count", None)
        if count is not None:
            gc_responses = list(self.gcode_queue)[-count:]
        else:
            gc_responses = list(self.gcode_queue)
        return {'gcode_store': gc_responses}

def load_component(config: ConfigHelper) -> DataStore:
    return DataStore(config)



# Helper for Moonraker to Klippy API calls.
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import logging
from ..utils import Sentinel
from ..common import WebRequest, APITransport, RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    Optional,
    Dict,
    List,
    TypeVar,
    Mapping,
    Callable,
    Coroutine
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import UserInfo
    from .klippy_connection import KlippyConnection as Klippy
    Subscription = Dict[str, Optional[List[Any]]]
    SubCallback = Callable[[Dict[str, Dict[str, Any]], float], Optional[Coroutine]]
    _T = TypeVar("_T")

INFO_ENDPOINT = "info"
ESTOP_ENDPOINT = "emergency_stop"
LIST_EPS_ENDPOINT = "list_endpoints"
GC_OUTPUT_ENDPOINT = "gcode/subscribe_output"
GCODE_ENDPOINT = "gcode/script"
SUBSCRIPTION_ENDPOINT = "objects/subscribe"
STATUS_ENDPOINT = "objects/query"
OBJ_LIST_ENDPOINT = "objects/list"
REG_METHOD_ENDPOINT = "register_remote_method"

class KlippyAPI(APITransport):
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.klippy: Klippy = self.server.lookup_component("klippy_connection")
        self.eventloop = self.server.get_event_loop()
        app_args = self.server.get_app_args()
        self.version = app_args.get('software_version')
        # Maintain a subscription for all moonraker requests, as
        # we do not want to overwrite them
        self.host_subscription: Subscription = {}
        self.subscription_callbacks: List[SubCallback] = []

        # Register GCode Aliases
        self.server.register_endpoint(
            "/printer/print/pause", RequestType.POST, self._gcode_pause
        )
        self.server.register_endpoint(
            "/printer/print/resume", RequestType.POST, self._gcode_resume
        )
        self.server.register_endpoint(
            "/printer/print/cancel", RequestType.POST, self._gcode_cancel
        )
        self.server.register_endpoint(
            "/printer/print/start", RequestType.POST, self._gcode_start_print
        )
        self.server.register_endpoint(
            "/printer/restart", RequestType.POST, self._gcode_restart
        )
        self.server.register_endpoint(
            "/printer/firmware_restart", RequestType.POST, self._gcode_firmware_restart
        )
        self.server.register_event_handler(
            "server:klippy_disconnect", self._on_klippy_disconnect
        )

    def _on_klippy_disconnect(self) -> None:
        self.host_subscription.clear()
        self.subscription_callbacks.clear()

    async def _gcode_pause(self, web_request: WebRequest) -> str:
        return await self.pause_print()

    async def _gcode_resume(self, web_request: WebRequest) -> str:
        return await self.resume_print()

    async def _gcode_cancel(self, web_request: WebRequest) -> str:
        return await self.cancel_print()

    async def _gcode_start_print(self, web_request: WebRequest) -> str:
        filename: str = web_request.get_str('filename')
        user = web_request.get_current_user()
        return await self.start_print(filename, user=user)

    async def _gcode_restart(self, web_request: WebRequest) -> str:
        return await self.do_restart("RESTART")

    async def _gcode_firmware_restart(self, web_request: WebRequest) -> str:
        return await self.do_restart("FIRMWARE_RESTART")

    async def _send_klippy_request(
        self,
        method: str,
        params: Dict[str, Any],
        default: Any = Sentinel.MISSING,
        transport: Optional[APITransport] = None
    ) -> Any:
        try:
            req = WebRequest(method, params, transport=transport or self)
            result = await self.klippy.request(req)
        except self.server.error:
            if default is Sentinel.MISSING:
                raise
            result = default
        return result

    async def run_gcode(self,
                        script: str,
                        default: Any = Sentinel.MISSING
                        ) -> str:
        params = {'script': script}
        result = await self._send_klippy_request(
            GCODE_ENDPOINT, params, default)
        return result

    async def start_print(
        self,
        filename: str,
        wait_klippy_started: bool = False,
        user: Optional[UserInfo] = None
    ) -> str:
        # WARNING: Do not call this method from within the following
        # event handlers when "wait_klippy_started" is set to True:
        # klippy_identified, klippy_started, klippy_ready, klippy_disconnect
        # Doing so will result in "wait_started" blocking for the specifed
        # timeout (default 20s) and returning False.
        # XXX - validate that file is on disk
        if filename[0] == '/':
            filename = filename[1:]
        # Escape existing double quotes in the file name
        filename = filename.replace("\"", "\\\"")
        script = f'SDCARD_PRINT_FILE FILENAME="{filename}"'
        if wait_klippy_started:
            await self.klippy.wait_started()
        logging.info(f"Requesting Job Start, filename = {filename}")
        ret = await self.run_gcode(script)
        self.server.send_event("klippy_apis:job_start_complete", user)
        return ret

    async def pause_print(
        self, default: Union[Sentinel, _T] = Sentinel.MISSING
    ) -> Union[_T, str]:
        self.server.send_event("klippy_apis:pause_requested")
        logging.info("Requesting job pause...")
        return await self._send_klippy_request(
            "pause_resume/pause", {}, default)

    async def resume_print(
        self, default: Union[Sentinel, _T] = Sentinel.MISSING
    ) -> Union[_T, str]:
        self.server.send_event("klippy_apis:resume_requested")
        logging.info("Requesting job resume...")
        return await self._send_klippy_request(
            "pause_resume/resume", {}, default)

    async def cancel_print(
        self, default: Union[Sentinel, _T] = Sentinel.MISSING
    ) -> Union[_T, str]:
        self.server.send_event("klippy_apis:cancel_requested")
        logging.info("Requesting job cancel...")
        return await self._send_klippy_request(
            "pause_resume/cancel", {}, default)

    async def do_restart(
        self, gc: str, wait_klippy_started: bool = False
    ) -> str:
        # WARNING: Do not call this method from within the following
        # event handlers when "wait_klippy_started" is set to True:
        # klippy_identified, klippy_started, klippy_ready, klippy_disconnect
        # Doing so will result in "wait_started" blocking for the specifed
        # timeout (default 20s) and returning False.
        if wait_klippy_started:
            await self.klippy.wait_started()
        try:
            result = await self.run_gcode(gc)
        except self.server.error as e:
            if str(e) == "Klippy Disconnected":
                result = "ok"
            else:
                raise
        return result

    async def list_endpoints(self,
                             default: Union[Sentinel, _T] = Sentinel.MISSING
                             ) -> Union[_T, Dict[str, List[str]]]:
        return await self._send_klippy_request(
            LIST_EPS_ENDPOINT, {}, default)

    async def emergency_stop(self) -> str:
        return await self._send_klippy_request(ESTOP_ENDPOINT, {})

    async def get_klippy_info(self,
                              send_id: bool = False,
                              default: Union[Sentinel, _T] = Sentinel.MISSING
                              ) -> Union[_T, Dict[str, Any]]:
        params = {}
        if send_id:
            ver = self.version
            params = {'client_info': {'program': "Moonraker", 'version': ver}}
        return await self._send_klippy_request(INFO_ENDPOINT, params, default)

    async def get_object_list(self,
                              default: Union[Sentinel, _T] = Sentinel.MISSING
                              ) -> Union[_T, List[str]]:
        result = await self._send_klippy_request(
            OBJ_LIST_ENDPOINT, {}, default)
        if isinstance(result, dict) and 'objects' in result:
            return result['objects']
        if default is not Sentinel.MISSING:
            return default
        raise self.server.error("Invalid response received from Klippy", 500)

    async def query_objects(self,
                            objects: Mapping[str, Optional[List[str]]],
                            default: Union[Sentinel, _T] = Sentinel.MISSING
                            ) -> Union[_T, Dict[str, Any]]:
        params = {'objects': objects}
        result = await self._send_klippy_request(
            STATUS_ENDPOINT, params, default)
        if isinstance(result, dict) and "status" in result:
            return result["status"]
        if default is not Sentinel.MISSING:
            return default
        raise self.server.error("Invalid response received from Klippy", 500)

    async def subscribe_objects(
        self,
        objects: Mapping[str, Optional[List[str]]],
        callback: Optional[SubCallback] = None,
        default: Union[Sentinel, _T] = Sentinel.MISSING
    ) -> Union[_T, Dict[str, Any]]:
        # The host transport shares subscriptions amongst all components
        for obj, items in objects.items():
            if obj in self.host_subscription:
                prev = self.host_subscription[obj]
                if items is None or prev is None:
                    self.host_subscription[obj] = None
                else:
                    uitems = list(set(prev) | set(items))
                    self.host_subscription[obj] = uitems
            else:
                self.host_subscription[obj] = items
        params = {"objects": dict(self.host_subscription)}
        result = await self._send_klippy_request(SUBSCRIPTION_ENDPOINT, params, default)
        if isinstance(result, dict) and "status" in result:
            if callback is not None:
                self.subscription_callbacks.append(callback)
            return result["status"]
        if default is not Sentinel.MISSING:
            return default
        raise self.server.error("Invalid response received from Klippy", 500)

    async def subscribe_from_transport(
        self,
        objects: Mapping[str, Optional[List[str]]],
        transport: APITransport,
        default: Union[Sentinel, _T] = Sentinel.MISSING,
    ) -> Union[_T, Dict[str, Any]]:
        params = {"objects": dict(objects)}
        result = await self._send_klippy_request(
            SUBSCRIPTION_ENDPOINT, params, default, transport
        )
        if isinstance(result, dict) and "status" in result:
            return result["status"]
        if default is not Sentinel.MISSING:
            return default
        raise self.server.error("Invalid response received from Klippy", 500)

    async def subscribe_gcode_output(self) -> str:
        template = {'response_template':
                    {'method': "process_gcode_response"}}
        return await self._send_klippy_request(GC_OUTPUT_ENDPOINT, template)

    async def register_method(self, method_name: str) -> str:
        return await self._send_klippy_request(
            REG_METHOD_ENDPOINT,
            {'response_template': {"method": method_name},
             'remote_method': method_name})

    def send_status(
        self, status: Dict[str, Any], eventtime: float
    ) -> None:
        for cb in self.subscription_callbacks:
            self.eventloop.register_callback(cb, status, eventtime)
        self.server.send_event("server:status_update", status)

def load_component(config: ConfigHelper) -> KlippyAPI:
    return KlippyAPI(config)



# Klipper Web Server Rest API
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import mimetypes
import logging
import traceback
import ssl
import pathlib
import urllib.parse
import tornado
import tornado.iostream
import tornado.httputil
import tornado.web
from asyncio import Lock
from inspect import isclass
from tornado.escape import url_unescape, url_escape
from tornado.routing import Rule, PathMatches, RuleRouter
from tornado.http1connection import HTTP1Connection
from tornado.httpserver import HTTPServer
from tornado.log import access_log
from ..utils import ServerError, source_info, parse_ip_address
from ..common import (
    JsonRPC,
    WebRequest,
    APIDefinition,
    APITransport,
    TransportType,
    RequestType,
    KlippyState
)
from ..utils import json_wrapper as jsonw
from streaming_form_data import StreamingFormDataParser, ParseFailedException
from streaming_form_data.targets import FileTarget, ValueTarget, SHA256Target

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Callable,
    Coroutine,
    Union,
    Dict,
    List,
    AsyncGenerator,
    Type
)
if TYPE_CHECKING:
    from tornado.websocket import WebSocketHandler
    from tornado.httputil import HTTPMessageDelegate, HTTPServerRequest
    from ..server import Server
    from ..eventloop import EventLoop
    from ..confighelper import ConfigHelper
    from ..common import UserInfo
    from .klippy_connection import KlippyConnection as Klippy
    from ..utils import IPAddress
    from .websockets import WebsocketManager, WebSocket
    from .file_manager.file_manager import FileManager
    from .announcements import Announcements
    from .machine import Machine
    from io import BufferedReader
    from .authorization import Authorization
    from .template import TemplateFactory, JinjaTemplate
    MessageDelgate = Optional[HTTPMessageDelegate]
    AuthComp = Optional[Authorization]
    APICallback = Callable[[WebRequest], Coroutine]

# mypy: disable-error-code="attr-defined,name-defined"

# 50 MiB Max Standard Body Size
MAX_BODY_SIZE = 50 * 1024 * 1024
MAX_WS_CONNS_DEFAULT = 50
EXCLUDED_ARGS = ["_", "token", "access_token", "connection_id"]
AUTHORIZED_EXTS = [".png", ".jpg"]
DEFAULT_KLIPPY_LOG_PATH = "/tmp/klippy.log"

class MutableRouter(RuleRouter):
    def __init__(self, application: tornado.web.Application) -> None:
        self.application = application
        self.pattern_to_rule: Dict[str, Rule] = {}
        super(MutableRouter, self).__init__(None)

    def get_target_delegate(self,
                            target: Any,
                            request: tornado.httputil.HTTPServerRequest,
                            **target_params
                            ) -> MessageDelgate:
        if isclass(target) and issubclass(target, tornado.web.RequestHandler):
            return self.application.get_handler_delegate(
                request, target, **target_params
            )
        return super(MutableRouter, self).get_target_delegate(
            target, request, **target_params)

    def has_rule(self, pattern: str) -> bool:
        return pattern in self.pattern_to_rule

    def add_handler(self,
                    pattern: str,
                    target: Any,
                    target_params: Optional[Dict[str, Any]] = None
                    ) -> None:
        if pattern in self.pattern_to_rule:
            self.remove_handler(pattern)
        new_rule = Rule(PathMatches(pattern), target, target_params)
        self.pattern_to_rule[pattern] = new_rule
        self.rules.append(new_rule)

    def remove_handler(self, pattern: str) -> None:
        rule = self.pattern_to_rule.pop(pattern, None)
        if rule is not None:
            try:
                self.rules.remove(rule)
            except Exception:
                logging.exception(f"Unable to remove rule: {pattern}")

class PrimaryRouter(MutableRouter):
    def __init__(self, config: ConfigHelper) -> None:
        server = config.get_server()
        max_ws_conns = config.getint('max_websocket_connections', MAX_WS_CONNS_DEFAULT)
        self.verbose_logging = server.is_verbose_enabled()
        app_args: Dict[str, Any] = {
            'serve_traceback': self.verbose_logging,
            'websocket_ping_interval': 10,
            'websocket_ping_timeout': 30,
            'server': server,
            'max_websocket_connections': max_ws_conns,
            'log_function': self.log_request
        }
        super().__init__(tornado.web.Application(**app_args))

    @property
    def tornado_app(self) -> tornado.web.Application:
        return self.application

    def find_handler(
        self, request: HTTPServerRequest, **kwargs: Any
    ) -> Optional[HTTPMessageDelegate]:
        hdlr = super().find_handler(request, **kwargs)
        if hdlr is not None:
            return hdlr
        return self.application.get_handler_delegate(request, AuthorizedErrorHandler)

    def log_request(self, handler: tornado.web.RequestHandler) -> None:
        status_code = handler.get_status()
        if (
            not self.verbose_logging and
            status_code in [200, 204, 206, 304]
        ):
            # don't log successful requests in release mode
            return
        if status_code < 400:
            log_method = access_log.info
        elif status_code < 500:
            log_method = access_log.warning
        else:
            log_method = access_log.error
        request_time = 1000.0 * handler.request.request_time()
        user: Optional[UserInfo] = handler.current_user
        username = "No User"
        if user is not None:
            username = user.username
        log_method(
            f"{status_code} {handler._request_summary()} "
            f"[{username}] {request_time:.2f}ms"
        )

class InternalTransport(APITransport):
    def __init__(self, server: Server) -> None:
        self.server = server

    async def call_method(self,
                          method_name: str,
                          request_arguments: Dict[str, Any] = {},
                          **kwargs
                          ) -> Any:
        rpc: JsonRPC = self.server.lookup_component("jsonrpc")
        method_info = rpc.get_method(method_name)
        if method_info is None:
            raise self.server.error(f"No method {method_name} available")
        req_type, api_definition = method_info
        if TransportType.INTERNAL not in api_definition.transports:
            raise self.server.error(f"No method {method_name} available")
        args = request_arguments or kwargs
        return await api_definition.request(args, req_type, self)

class MoonrakerApp:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.json_rpc = JsonRPC(self.server)
        self.http_server: Optional[HTTPServer] = None
        self.secure_server: Optional[HTTPServer] = None
        self.template_cache: Dict[str, JinjaTemplate] = {}
        self.registered_base_handlers: List[str] = [
            "/server/redirect",
            "/server/jsonrpc"
        ]
        self.max_upload_size = config.getint('max_upload_size', 1024)
        self.max_upload_size *= 1024 * 1024

        # SSL config
        self.cert_path: pathlib.Path = self._get_path_option(
            config, 'ssl_certificate_path')
        self.key_path: pathlib.Path = self._get_path_option(
            config, 'ssl_key_path')

        # Route Prefix
        home_pattern = "/"
        self._route_prefix: str = ""
        route_prefix = config.get("route_prefix", None)
        if route_prefix is not None:
            rparts = route_prefix.strip("/").split("/")
            rp = "/".join(
                [url_escape(part, plus=False) for part in rparts if part]
            )
            if not rp:
                raise config.error(
                    f"Invalid value for option 'route_prefix': {route_prefix}"
                )
            self._route_prefix = f"/{rp}"
            home_pattern = f"{self._route_prefix}/?"
        self.internal_transport = InternalTransport(self.server)

        mimetypes.add_type('text/plain', '.log')
        mimetypes.add_type('text/plain', '.gcode')
        mimetypes.add_type('text/plain', '.cfg')

        # Set up HTTP routing.  Our "mutable_router" wraps a Tornado Application
        self.mutable_router = PrimaryRouter(config)
        for (ptrn, hdlr) in (
            (home_pattern, WelcomeHandler),
            (f"{self._route_prefix}/server/redirect", RedirectHandler),
            (f"{self._route_prefix}/server/jsonrpc", RPCHandler)
        ):
            self.mutable_router.add_handler(ptrn, hdlr, None)

        # Register handlers
        logfile = self.server.get_app_args().get('log_file')
        if logfile:
            self.register_static_file_handler(
                "moonraker.log", logfile, force=True)
        self.register_static_file_handler(
            "klippy.log", DEFAULT_KLIPPY_LOG_PATH, force=True)
        self.register_upload_handler("/server/files/upload")

        # Register Server Components
        self.server.register_component("jsonrpc", self.json_rpc)
        self.server.register_component("internal_transport", self.internal_transport)

    def _get_path_option(
        self, config: ConfigHelper, option: str
    ) -> pathlib.Path:
        path: Optional[str] = config.get(option, None, deprecate=True)
        app_args = self.server.get_app_args()
        data_path = app_args["data_path"]
        certs_path = pathlib.Path(data_path).joinpath("certs")
        if not certs_path.exists():
            try:
                certs_path.mkdir()
            except Exception:
                pass
        ext = "key" if "key" in option else "cert"
        item = certs_path.joinpath(f"moonraker.{ext}")
        if item.exists() or path is None:
            return item
        item = pathlib.Path(path).expanduser().resolve()
        if not item.exists():
            raise self.server.error(
                f"Invalid path for option '{option}', "
                f"{path} does not exist"
            )
        return item

    @property
    def route_prefix(self):
        return self._route_prefix

    def parse_endpoint(self, http_path: str) -> str:
        if not self._route_prefix or not http_path.startswith(self._route_prefix):
            return http_path
        return http_path[len(self._route_prefix):]

    def listen(self, host: str, port: int, ssl_port: int) -> None:
        if host.lower() == "all":
            host = ""
        self.http_server = self._create_http_server(port, host)
        if self.https_enabled():
            if port == ssl_port:
                self.server.add_warning(
                    "Failed to start HTTPS server.  Server options 'port' and "
                    f"'ssl_port' match, both set to {port}.  Modify the "
                    "configuration to use different ports."
                )
                return
            logging.info(f"Starting secure server on port {ssl_port}")
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(self.cert_path, self.key_path)
            self.secure_server = self._create_http_server(
                ssl_port, host, ssl_options=ssl_ctx
            )
        else:
            logging.info(
                "SSL Certificate/Key not configured, aborting HTTPS Server startup"
            )

    def _create_http_server(
        self, port: int, address: str, **kwargs
    ) -> Optional[HTTPServer]:
        args: Dict[str, Any] = dict(max_body_size=MAX_BODY_SIZE, xheaders=True)
        args.update(kwargs)
        svr = HTTPServer(self.mutable_router, **args)
        try:
            svr.listen(port, address)
        except Exception as e:
            svr_type = "HTTPS" if "ssl_options" in args else "HTTP"
            self.server.add_warning(
                f"Failed to start {svr_type} server: {e}.  See moonraker.log "
                "for more details.", exc_info=e
            )
            return None
        return svr

    def get_server(self) -> Server:
        return self.server

    def https_enabled(self) -> bool:
        return self.cert_path.exists() and self.key_path.exists()

    async def close(self) -> None:
        if self.http_server is not None:
            self.http_server.stop()
            await self.http_server.close_all_connections()
        if self.secure_server is not None:
            self.secure_server.stop()
            await self.secure_server.close_all_connections()
        APIDefinition.reset_cache()

    def register_endpoint(
        self,
        endpoint: str,
        request_types: Union[List[str], RequestType],
        callback: APICallback,
        transports: Union[List[str], TransportType] = TransportType.all(),
        wrap_result: bool = True,
        content_type: Optional[str] = None,
        auth_required: bool = True,
        is_remote: bool = False
    ) -> None:
        if isinstance(request_types, list):
            request_types = RequestType.from_string_list(request_types)
        if isinstance(transports, list):
            transports = TransportType.from_string_list(transports)
        api_def = APIDefinition.create(
            endpoint, request_types, callback, transports, auth_required, is_remote
        )
        http_path = api_def.http_path
        if http_path in self.registered_base_handlers:
            if not is_remote:
                raise self.server.error(
                    f"Local endpoint '{endpoint}' already registered"
                )
            return
        logging.debug(f"Registering API: {api_def}")
        if TransportType.HTTP in transports:
            params: dict[str, Any] = {}
            params["api_definition"] = api_def
            params["wrap_result"] = wrap_result
            params["content_type"] = content_type
            self.mutable_router.add_handler(
                f"{self._route_prefix}{http_path}", DynamicRequestHandler, params
            )
        self.registered_base_handlers.append(http_path)
        for request_type, method_name in api_def.rpc_items():
            self.json_rpc.register_method(method_name, request_type, api_def)

    def register_static_file_handler(
        self, pattern: str, file_path: str, force: bool = False
    ) -> None:
        if pattern[0] != "/":
            pattern = "/server/files/" + pattern
        if os.path.isfile(file_path) or force:
            pattern += '()'
        elif os.path.isdir(file_path):
            if pattern[-1] != "/":
                pattern += "/"
            pattern += "(.*)"
        else:
            logging.info(f"Invalid file path: {file_path}")
            return
        logging.debug(f"Registering static file: ({pattern}) {file_path}")
        params = {'path': file_path}
        self.mutable_router.add_handler(
            f"{self._route_prefix}{pattern}", FileRequestHandler, params
        )

    def register_upload_handler(
        self, pattern: str, location_prefix: str = "server/files"
    ) -> None:
        params: Dict[str, Any] = {'max_upload_size': self.max_upload_size}
        location_prefix = location_prefix.strip("/")
        if self._route_prefix:
            location_prefix = f"{self._route_prefix.strip('/')}/{location_prefix}"
        params['location_prefix'] = location_prefix
        self.mutable_router.add_handler(
            f"{self._route_prefix}{pattern}", FileUploadHandler, params
        )

    def register_websocket_handler(
        self, pattern: str, handler: Type[WebSocketHandler]
    ) -> None:
        self.mutable_router.add_handler(
            f"{self._route_prefix}{pattern}", handler, None
        )

    def register_debug_endpoint(
        self,
        endpoint: str,
        request_types: Union[List[str], RequestType],
        callback: APICallback,
        transports: Union[List[str], TransportType] = TransportType.all(),
        wrap_result: bool = True
    ) -> None:
        if not self.server.is_debug_enabled():
            return
        if not endpoint.startswith("/debug"):
            raise self.server.error(
                "Debug Endpoints must be registered in the '/debug' path"
            )
        self.register_endpoint(
            endpoint, request_types, callback, transports, wrap_result
        )

    def remove_endpoint(self, endpoint: str) -> None:
        api_def = APIDefinition.pop_cached_def(endpoint)
        if api_def is not None:
            logging.debug(f"Removing Endpoint: {endpoint}")
            if api_def.http_path in self.registered_base_handlers:
                self.registered_base_handlers.remove(api_def.http_path)
            self.mutable_router.remove_handler(api_def.http_path)
            for method_name in api_def.rpc_methods:
                self.json_rpc.remove_method(method_name)

    async def load_template(self, asset_name: str) -> JinjaTemplate:
        if asset_name in self.template_cache:
            return self.template_cache[asset_name]
        eventloop = self.server.get_event_loop()
        asset = await eventloop.run_in_thread(
            source_info.read_asset, asset_name
        )
        if asset is None:
            raise tornado.web.HTTPError(404, "Asset Not Found")
        template: TemplateFactory = self.server.lookup_component("template")
        asset_tmpl = template.create_ui_template(asset)
        self.template_cache[asset_name] = asset_tmpl
        return asset_tmpl

def _set_cors_headers(req_hdlr: tornado.web.RequestHandler) -> None:
    request = req_hdlr.request
    origin: Optional[str] = request.headers.get("Origin")
    if origin is None:
        return
    req_hdlr.set_header("Access-Control-Allow-Origin", origin)
    if req_hdlr.request.method == "OPTIONS":
        req_hdlr.set_header(
            "Access-Control-Allow-Methods",
            "GET, POST, PUT, DELETE, OPTIONS"
        )
        req_hdlr.set_header(
            "Access-Control-Allow-Headers",
            "Origin, Accept, Content-Type, X-Requested-With, "
            "X-CRSF-Token, Authorization, X-Access-Token, "
            "X-Api-Key"
        )
        req_pvt_header = req_hdlr.request.headers.get(
            "Access-Control-Request-Private-Network", None
        )
        if req_pvt_header == "true":
            req_hdlr.set_header("Access-Control-Allow-Private-Network", "true")


class AuthorizedRequestHandler(tornado.web.RequestHandler):
    def initialize(self) -> None:
        self.server: Server = self.settings['server']
        self.auth_required: bool = True
        self.cors_enabled = False

    def set_default_headers(self) -> None:
        if getattr(self, "cors_enabled", False):
            _set_cors_headers(self)

    async def prepare(self) -> None:
        auth: AuthComp = self.server.lookup_component('authorization', None)
        if auth is not None:
            origin: Optional[str] = self.request.headers.get("Origin")
            self.cors_enabled = await auth.check_cors(origin)
            if self.cors_enabled:
                _set_cors_headers(self)
            self.current_user = await auth.authenticate_request(
                self.request, self.auth_required
            )

    def options(self, *args, **kwargs) -> None:
        # Enable CORS if configured
        if self.cors_enabled:
            self.set_status(204)
            self.finish()
        else:
            super(AuthorizedRequestHandler, self).options()

    def get_associated_websocket(self) -> Optional[WebSocket]:
        # Return associated websocket connection if an id
        # was provided by the request
        conn = None
        conn_id: Any = self.get_argument('connection_id', None)
        if conn_id is not None:
            try:
                conn_id = int(conn_id)
            except Exception:
                pass
            else:
                wsm: WebsocketManager = self.server.lookup_component("websockets")
                conn = wsm.get_client_ws(conn_id)
        return conn

    def write_error(self, status_code: int, **kwargs) -> None:
        err = {'code': status_code, 'message': self._reason}
        if 'exc_info' in kwargs:
            err['traceback'] = "\n".join(
                traceback.format_exception(*kwargs['exc_info']))
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(jsonw.dumps({'error': err}))

# Due to the way Python treats multiple inheritance its best
# to create a separate authorized handler for serving files
class AuthorizedFileHandler(tornado.web.StaticFileHandler):
    def initialize(self,
                   path: str,
                   default_filename: Optional[str] = None
                   ) -> None:
        super(AuthorizedFileHandler, self).initialize(path, default_filename)
        self.server: Server = self.settings['server']
        self.cors_enabled = False

    def set_default_headers(self) -> None:
        if getattr(self, "cors_enabled", False):
            _set_cors_headers(self)

    async def prepare(self) -> None:
        auth: AuthComp = self.server.lookup_component('authorization', None)
        if auth is not None:
            origin: Optional[str] = self.request.headers.get("Origin")
            self.cors_enabled = await auth.check_cors(origin)
            if self.cors_enabled:
                _set_cors_headers(self)
            self.current_user = await auth.authenticate_request(
                self.request, self._check_need_auth()
            )

    def options(self, *args, **kwargs) -> None:
        # Enable CORS if configured
        if self.cors_enabled:
            self.set_status(204)
            self.finish()
        else:
            super(AuthorizedFileHandler, self).options()

    def write_error(self, status_code: int, **kwargs) -> None:
        err = {'code': status_code, 'message': self._reason}
        if 'exc_info' in kwargs:
            err['traceback'] = "\n".join(
                traceback.format_exception(*kwargs['exc_info']))
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(jsonw.dumps({'error': err}))

    def _check_need_auth(self) -> bool:
        if self.request.method != "GET":
            return True
        ext = os.path.splitext(self.request.path)[-1].lower()
        if ext in AUTHORIZED_EXTS:
            return False
        return True

class DynamicRequestHandler(AuthorizedRequestHandler):
    def initialize(
        self,
        api_definition: Optional[APIDefinition] = None,
        wrap_result: bool = True,
        content_type: Optional[str] = None
    ) -> None:
        super(DynamicRequestHandler, self).initialize()
        assert api_definition is not None
        self.api_defintion = api_definition
        self.wrap_result = wrap_result
        self.content_type = content_type
        self.auth_required = api_definition.auth_required

    # Converts query string values with type hints
    def _convert_type(self, value: str, hint: str) -> Any:
        type_funcs: Dict[str, Callable] = {
            "int": int, "float": float,
            "bool": lambda x: x.lower() == "true",
            "json": jsonw.loads}
        if hint not in type_funcs:
            logging.info(f"No conversion method for type hint {hint}")
            return value
        func = type_funcs[hint]
        try:
            converted = func(value)
        except Exception:
            logging.exception("Argument conversion error: Hint: "
                              f"{hint}, Arg: {value}")
            return value
        return converted

    def _default_parser(self) -> Dict[str, Any]:
        args = {}
        for key in self.request.arguments.keys():
            if key in EXCLUDED_ARGS:
                continue
            key_parts = key.rsplit(":", 1)
            val = self.get_argument(key)
            if len(key_parts) == 1:
                args[key] = val
            else:
                args[key_parts[0]] = self._convert_type(val, key_parts[1])
        return args

    def _object_parser(self) -> Dict[str, Dict[str, Any]]:
        args: Dict[str, Any] = {}
        for key in self.request.arguments.keys():
            if key in EXCLUDED_ARGS:
                continue
            val = self.get_argument(key)
            if not val:
                args[key] = None
            else:
                args[key] = val.split(',')
        logging.debug(f"Parsed Arguments: {args}")
        return {'objects': args}

    def parse_args(self) -> Dict[str, Any]:
        try:
            if self.api_defintion.need_object_parser:
                args: Dict[str, Any] = self._object_parser()
            else:
                args = self._default_parser()
        except Exception:
            raise ServerError(
                "Error Parsing Request Arguments. "
                "Is the Content-Type correct?")
        content_type = self.request.headers.get('Content-Type', "").strip()
        if content_type.startswith("application/json"):
            try:
                args.update(jsonw.loads(self.request.body))
            except jsonw.JSONDecodeError:
                pass
        for key, value in self.path_kwargs.items():
            if value is not None:
                args[key] = value
        return args

    def _log_debug(self, header: str, args: Any) -> None:
        if self.server.is_verbose_enabled():
            resp = args
            endpoint = self.api_defintion.endpoint
            if isinstance(args, dict):
                if (
                    endpoint.startswith("/access") or
                    endpoint.startswith("/machine/sudo/password")
                ):
                    resp = {key: "<sanitized>" for key in args}
            elif isinstance(args, str):
                if args.startswith("<html>"):
                    resp = "<html>"
            logging.debug(f"{header}::{resp}")

    async def get(self, *args, **kwargs) -> None:
        await self._process_http_request(RequestType.GET)

    async def post(self, *args, **kwargs) -> None:
        await self._process_http_request(RequestType.POST)

    async def delete(self, *args, **kwargs) -> None:
        await self._process_http_request(RequestType.DELETE)

    async def _process_http_request(self, req_type: RequestType) -> None:
        if req_type not in self.api_defintion.request_types:
            raise tornado.web.HTTPError(405)
        args = self.parse_args()
        transport = self.get_associated_websocket()
        req = f"{self.request.method} {self.request.path}"
        self._log_debug(f"HTTP Request::{req}", args)
        try:
            ip = parse_ip_address(self.request.remote_ip or "")
            result = await self.api_defintion.request(
                args, req_type, transport, ip, self.current_user
            )
        except ServerError as e:
            raise tornado.web.HTTPError(
                e.status_code, reason=str(e)) from e
        if self.wrap_result:
            result = {'result': result}
        self._log_debug(f"HTTP Response::{req}", result)
        if result is None:
            self.set_status(204)
        elif isinstance(result, dict):
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            result = jsonw.dumps(result)
        elif self.content_type is not None:
            self.set_header("Content-Type", self.content_type)
        self.finish(result)

class RPCHandler(AuthorizedRequestHandler, APITransport):
    def initialize(self) -> None:
        super(RPCHandler, self).initialize()
        self.auth_required = False

    @property
    def transport_type(self) -> TransportType:
        return TransportType.HTTP

    @property
    def user_info(self) -> Optional[UserInfo]:
        return self.current_user

    @property
    def ip_addr(self) -> Optional[IPAddress]:
        return parse_ip_address(self.request.remote_ip or "")

    def screen_rpc_request(
        self, api_def: APIDefinition, req_type: RequestType, args: Dict[str, Any]
    ) -> None:
        if self.current_user is None and api_def.auth_required:
            raise self.server.error("Unauthorized", 401)
        if api_def.endpoint == "objects/subscribe":
            raise self.server.error(
                "Subscriptions not available for HTTP transport", 404
            )

    def send_status(self, status: Dict[str, Any], eventtime: float) -> None:
        # Can't handle status updates.  This should not be called, but
        # we don't want to raise an exception if it is
        pass

    async def post(self, *args, **kwargs) -> None:
        content_type = self.request.headers.get('Content-Type', "").strip()
        if not content_type.startswith("application/json"):
            raise tornado.web.HTTPError(
                400, "Invalid content type, application/json required"
            )
        rpc: JsonRPC = self.server.lookup_component("jsonrpc")
        result = await rpc.dispatch(self.request.body, self)
        if result is not None:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(result)

class FileRequestHandler(AuthorizedFileHandler):
    def set_extra_headers(self, path: str) -> None:
        # The call below shold never return an empty string,
        # as the path should have already been validated to be
        # a file
        assert isinstance(self.absolute_path, str)
        basename = os.path.basename(self.absolute_path)
        ascii_basename = self._escape_filename_to_ascii(basename)
        utf8_basename = self._escape_filename_to_utf8(basename)
        self.set_header(
            "Content-Disposition",
            f"attachment; filename=\"{ascii_basename}\"; "
            f"filename*=UTF-8\'\'{utf8_basename}")

    async def delete(self, path: str) -> None:
        app: MoonrakerApp = self.server.lookup_component("application")
        endpoint = app.parse_endpoint(self.request.path or "")
        path = endpoint.lstrip("/").split("/", 2)[-1]
        path = url_unescape(path, plus=False)
        file_manager: FileManager
        file_manager = self.server.lookup_component('file_manager')
        try:
            filename = await file_manager.delete_file(path)
        except self.server.error as e:
            raise tornado.web.HTTPError(e.status_code, str(e))
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(jsonw.dumps({'result': filename}))

    async def get(self, path: str, include_body: bool = True) -> None:
        # Set up our path instance variables.
        self.path = self.parse_url_path(path)
        del path  # make sure we don't refer to path instead of self.path again
        absolute_path = self.get_absolute_path(self.root, self.path)
        self.absolute_path = self.validate_absolute_path(
            self.root, absolute_path)
        if self.absolute_path is None:
            return
        file_manager: FileManager
        file_manager = self.server.lookup_component('file_manager')
        try:
            file_manager.check_reserved_path(self.absolute_path, False)
        except self.server.error as e:
            raise tornado.web.HTTPError(e.status_code, str(e))

        self.modified = self.get_modified_time()
        self.set_headers()

        self.request.headers.pop("If-None-Match", None)
        if self.should_return_304():
            self.set_status(304)
            return

        request_range = None
        range_header = self.request.headers.get("Range")
        if range_header:
            # As per RFC 2616 14.16, if an invalid Range header is specified,
            # the request will be treated as if the header didn't exist.
            request_range = tornado.httputil._parse_request_range(range_header)

        size = self.get_content_size()
        if request_range:
            start, end = request_range
            if start is not None and start < 0:
                start += size
                if start < 0:
                    start = 0
            if (
                start is not None
                and (start >= size or (end is not None and start >= end))
            ) or end == 0:
                # As per RFC 2616 14.35.1, a range is not satisfiable only: if
                # the first requested byte is equal to or greater than the
                # content, or when a suffix with length 0 is specified.
                # https://tools.ietf.org/html/rfc7233#section-2.1
                # A byte-range-spec is invalid if the last-byte-pos value is
                # present and less than the first-byte-pos.
                self.set_status(416)  # Range Not Satisfiable
                self.set_header("Content-Type", "text/plain")
                self.set_header("Content-Range", "bytes */%s" % (size,))
                return
            if end is not None and end > size:
                # Clients sometimes blindly use a large range to limit their
                # download size; cap the endpoint at the actual file size.
                end = size
            # Note: only return HTTP 206 if less than the entire range has been
            # requested. Not only is this semantically correct, but Chrome
            # refuses to play audio if it gets an HTTP 206 in response to
            # ``Range: bytes=0-``.
            if size != (end or size) - (start or 0):
                self.set_status(206)  # Partial Content
                self.set_header(
                    "Content-Range", tornado.httputil._get_content_range(
                        start, end, size)
                )
        else:
            start = end = None

        if start is not None and end is not None:
            content_length = end - start
        elif end is not None:
            content_length = end
        elif start is not None:
            end = size
            content_length = size - start
        else:
            end = size
            content_length = size
        self.set_header("Content-Length", content_length)

        if include_body:
            evt_loop = self.server.get_event_loop()
            content = self.get_content_nonblock(
                evt_loop, self.absolute_path, start, end)
            async for chunk in content:
                try:
                    self.write(chunk)
                    await self.flush()
                except tornado.iostream.StreamClosedError:
                    return
        else:
            assert self.request.method == "HEAD"

    def _escape_filename_to_ascii(self, basename: str) -> str:
        ret = basename.encode("ascii", "replace").decode()
        return ret.replace('"', '\\"')

    def _escape_filename_to_utf8(self, basename: str) -> str:
        return urllib.parse.quote(basename, encoding="utf-8")

    @classmethod
    async def get_content_nonblock(
        cls,
        evt_loop: EventLoop,
        abspath: str,
        start: Optional[int] = None,
        end: Optional[int] = None
    ) -> AsyncGenerator[bytes, None]:
        file: BufferedReader = await evt_loop.run_in_thread(open, abspath, "rb")
        try:
            if start is not None:
                file.seek(start)
            if end is not None:
                remaining = end - (start or 0)  # type: Optional[int]
            else:
                remaining = None
            while True:
                chunk_size = 64 * 1024
                if remaining is not None and remaining < chunk_size:
                    chunk_size = remaining
                chunk = await evt_loop.run_in_thread(file.read, chunk_size)
                if chunk:
                    if remaining is not None:
                        remaining -= len(chunk)
                    yield chunk
                else:
                    if remaining is not None:
                        assert remaining == 0
                    return
        finally:
            await evt_loop.run_in_thread(file.close)

    @classmethod
    def _get_cached_version(cls, abs_path: str) -> Optional[str]:
        return None

@tornado.web.stream_request_body
class FileUploadHandler(AuthorizedRequestHandler):
    def initialize(self,
                   location_prefix: str = "server/files",
                   max_upload_size: int = MAX_BODY_SIZE
                   ) -> None:
        self.location_prefix = location_prefix
        super(FileUploadHandler, self).initialize()
        self.file_manager: FileManager = self.server.lookup_component(
            'file_manager')
        self.max_upload_size = max_upload_size
        self.parse_lock = Lock()
        self.parse_failed: bool = False

    async def prepare(self) -> None:
        ret = super(FileUploadHandler, self).prepare()
        if ret is not None:
            await ret
        content_type: str = self.request.headers.get("Content-Type", "")
        logging.info(
            f"Upload Request Received from {self.request.remote_ip}\n"
            f"Content-Type: {content_type}"
        )
        fm: FileManager = self.server.lookup_component("file_manager")
        fm.check_write_enabled()
        if self.request.method == "POST":
            assert isinstance(self.request.connection, HTTP1Connection)
            self.request.connection.set_max_body_size(self.max_upload_size)
            tmpname = self.file_manager.gen_temp_upload_path()
            self._targets = {
                'root': ValueTarget(),
                'print': ValueTarget(),
                'path': ValueTarget(),
                'checksum': ValueTarget(),
            }
            self._file = FileTarget(tmpname)
            self._sha256_target = SHA256Target()
            self._parser = StreamingFormDataParser(self.request.headers)
            self._parser.register('file', self._file)
            self._parser.register('file', self._sha256_target)
            for name, target in self._targets.items():
                self._parser.register(name, target)

    async def data_received(self, chunk: bytes) -> None:
        if self.request.method == "POST" and not self.parse_failed:
            async with self.parse_lock:
                evt_loop = self.server.get_event_loop()
                try:
                    await evt_loop.run_in_thread(self._parser.data_received, chunk)
                except ParseFailedException:
                    logging.exception("Chunk Parsing Error")
                    self.parse_failed = True

    async def post(self) -> None:
        if self.parse_failed:
            self._file.on_finish()
            try:
                os.remove(self._file.filename)
            except Exception:
                pass
            raise tornado.web.HTTPError(500, "File Upload Parsing Failed")
        form_args = {}
        chk_target = self._targets.pop('checksum')
        calc_chksum = self._sha256_target.value.lower()
        if chk_target.value:
            # Validate checksum
            recd_cksum = chk_target.value.decode().lower()
            if calc_chksum != recd_cksum:
                # remove temporary file
                try:
                    os.remove(self._file.filename)
                except Exception:
                    pass
                raise tornado.web.HTTPError(
                    422,
                    f"File checksum mismatch: expected {recd_cksum}, "
                    f"calculated {calc_chksum}"
                )
        for name, target in self._targets.items():
            if target.value:
                form_args[name] = target.value.decode()
        form_args['filename'] = self._file.multipart_filename
        form_args['tmp_file_path'] = self._file.filename
        debug_msg = "\nFile Upload Arguments:"
        for name, value in form_args.items():
            debug_msg += f"\n{name}: {value}"
        debug_msg += f"\nChecksum: {calc_chksum}"
        form_args["current_user"] = self.current_user
        logging.debug(debug_msg)
        logging.info(f"Processing Uploaded File: {self._file.multipart_filename}")
        try:
            result = await self.file_manager.finalize_upload(form_args)
        except ServerError as e:
            raise tornado.web.HTTPError(
                e.status_code, str(e))
        # Return 201 and add the Location Header
        item: Dict[str, Any] = result.get('item', {})
        root: Optional[str] = item.get('root', None)
        fpath: Optional[str] = item.get('path', None)
        if root is not None and fpath is not None:
            path_parts = fpath.split("/")
            fpath = "/".join([url_escape(p, plus=False) for p in path_parts])
            proto = self.request.protocol
            if not isinstance(proto, str):
                proto = "http"
            host = self.request.host
            if not isinstance(host, str):
                si = self.server.get_host_info()
                port = si['port'] if proto == "http" else si['ssl_port']
                host = f"{si['address']}:{port}"
            location = f"{proto}://{host}/{self.location_prefix}/{root}/{fpath}"
            self.set_header("Location", location)
            logging.debug(f"Upload Location header set: {location}")
        self.set_status(201)
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(jsonw.dumps(result))

# Default Handler for unregistered endpoints
class AuthorizedErrorHandler(AuthorizedRequestHandler):
    async def prepare(self) -> None:
        ret = super(AuthorizedErrorHandler, self).prepare()
        if ret is not None:
            await ret
        self.set_status(404)
        raise tornado.web.HTTPError(404)

    def check_xsrf_cookie(self) -> None:
        pass

    def write_error(self, status_code: int, **kwargs) -> None:
        err = {'code': status_code, 'message': self._reason}
        if 'exc_info' in kwargs:
            err['traceback'] = "\n".join(
                traceback.format_exception(*kwargs['exc_info']))
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.finish(jsonw.dumps({'error': err}))

class RedirectHandler(AuthorizedRequestHandler):
    def initialize(self) -> None:
        super().initialize()
        self.auth_required = False

    async def get(self, *args, **kwargs) -> None:
        url: Optional[str] = self.get_argument('url', None)
        if url is None:
            try:
                body_args: Dict[str, Any] = jsonw.loads(self.request.body)
            except jsonw.JSONDecodeError:
                body_args = {}
            if 'url' not in body_args:
                raise tornado.web.HTTPError(
                    400, "No url argument provided")
            url = body_args['url']
            assert url is not None
        # validate the url origin
        auth: AuthComp = self.server.lookup_component('authorization', None)
        if auth is None or not await auth.check_cors(url.rstrip("/")):
            raise tornado.web.HTTPError(
                400, f"Unauthorized URL redirect: {url}")
        self.redirect(url)

class WelcomeHandler(tornado.web.RequestHandler):
    def initialize(self) -> None:
        self.server: Server = self.settings['server']

    async def get(self) -> None:
        summary: List[str] = []
        auth: AuthComp = self.server.lookup_component("authorization", None)
        if auth is not None:
            try:
                await auth.authenticate_request(self.request)
            except tornado.web.HTTPError:
                authorized = False
            else:
                authorized = True
            if authorized:
                summary.append(
                    "Your device is authorized to access Moonraker's API."
                )
            else:
                summary.append(
                    "Your device is not authorized to access Moonraker's API. "
                    "This is normal if you intend to use API Key "
                    "authentication or log in as an authenticated user.  "
                    "Otherwise you need to add your IP address to the "
                    "'trusted_clients' option in the [authorization] section "
                    "of moonraker.conf."
                )
            cors_enabled = auth.cors_enabled()
            if cors_enabled:
                summary.append(
                    "CORS is enabled.  Cross origin requests will be allowed "
                    "for origins that match one of the patterns specified in "
                    "the 'cors_domain' option of the [authorization] section."
                )
            else:
                summary.append(
                    "All cross origin requests will be blocked by the browser. "
                    "The 'cors_domains' option in [authorization] must be  "
                    "configured to enable CORS."
                )
        else:
            authorized = True
            cors_enabled = False
            summary.append(
                "The [authorization] component is not enabled in "
                "moonraker.conf.  All connections will be considered trusted."
            )
            summary.append(
                "All cross origin requests will be blocked by the browser.  "
                "The [authorization] section in moonraker.conf must be "
                "configured to enable CORS."
            )
        kconn: Klippy = self.server.lookup_component("klippy_connection")
        kstate = kconn.state
        if kstate != KlippyState.DISCONNECTED:
            summary.append(f"Klipper reports {kstate.message.lower()}")
        else:
            summary.append(
                "Moonraker is not currently connected to Klipper.  Make sure "
                "that the klipper service has successfully started and that "
                "its unix is enabled."
            )
        ancomp: Announcements
        ancomp = self.server.lookup_component("announcements")
        wsm: WebsocketManager = self.server.lookup_component("websockets")
        machine: Machine = self.server.lookup_component("machine")
        svc_info = machine.get_moonraker_service_info()
        sudo_req_msg = "<br/>".join(machine.sudo_request_messages)
        context: Dict[str, Any] = {
            "remote_ip": self.request.remote_ip,
            "authorized": authorized,
            "cors_enabled": cors_enabled,
            "version": self.server.get_app_args()["software_version"],
            "ws_count": wsm.get_count(),
            "klippy_state": kstate,
            "warnings": self.server.get_warnings(),
            "summary": summary,
            "announcements": await ancomp.get_announcements(),
            "sudo_requested": machine.sudo_requested,
            "sudo_request_message": sudo_req_msg,
            "linux_user": machine.linux_user,
            "local_ip": machine.public_ip or "unknown",
            "service_name": svc_info.get("unit_name", "unknown"),
            "hostname": self.server.get_host_info()["hostname"],
        }
        app: MoonrakerApp = self.server.lookup_component("application")
        welcome_template = await app.load_template("welcome.html")
        ret = await welcome_template.render_async(context)
        self.finish(ret)

def load_component(config: ConfigHelper) -> MoonrakerApp:
    return MoonrakerApp(config)



# Sqlite database for Moonraker persistent storage
#
# Copyright (C) 2021-2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import pathlib
import struct
import operator
import inspect
import logging
import contextlib
import time
from asyncio import Future, Task, Lock
from functools import reduce
from queue import Queue
from threading import Thread
import sqlite3
from ..utils import Sentinel, ServerError
from ..utils import json_wrapper as jsonw
from ..common import RequestType, SqlTableDefinition

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    TypeVar,
    Tuple,
    Optional,
    Union,
    Dict,
    List,
    Set,
    Type,
    Sequence,
    Generator
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .klippy_connection import KlippyConnection
    from lmdb import Environment as LmdbEnvironment
    from types import TracebackType
    DBRecord = Optional[Union[int, float, bool, str, List[Any], Dict[str, Any]]]
    DBType = DBRecord
    SqlParams = Union[List[Any], Tuple[Any, ...], Dict[str, Any]]
    _T = TypeVar("_T")

DATABASE_VERSION = 2
SQL_DB_FILENAME = "moonraker-sql.db"
NAMESPACE_TABLE = "namespace_store"
REGISTRATION_TABLE = "table_registry"
SCHEMA_TABLE = (
    "sqlite_schema" if sqlite3.sqlite_version_info >= (3, 33, 0)
    else "sqlite_master"
)

RECORD_ENCODE_FUNCS: Dict[Type, Callable[..., bytes]] = {
    int: lambda x: b"q" + struct.pack("q", x),
    float: lambda x: b"d" + struct.pack("d", x),
    bool: lambda x: b"?" + struct.pack("?", x),
    str: lambda x: b"s" + x.encode(),
    list: lambda x: jsonw.dumps(x),
    dict: lambda x: jsonw.dumps(x),
    type(None): lambda x: b"\x00",
}

RECORD_DECODE_FUNCS: Dict[int, Callable[..., DBRecord]] = {
    ord("q"): lambda x: struct.unpack("q", x[1:])[0],
    ord("d"): lambda x: struct.unpack("d", x[1:])[0],
    ord("?"): lambda x: struct.unpack("?", x[1:])[0],
    ord("s"): lambda x: bytes(x[1:]).decode(),
    ord("["): lambda x: jsonw.loads(bytes(x)),
    ord("{"): lambda x: jsonw.loads(bytes(x)),
    0: lambda _: None
}

def encode_record(value: DBRecord) -> bytes:
    try:
        enc_func = RECORD_ENCODE_FUNCS[type(value)]
        return enc_func(value)
    except Exception:
        raise ServerError(
            f"Error encoding val: {value}, type: {type(value)}"
        )

def decode_record(bvalue: bytes) -> DBRecord:
    fmt = bvalue[0]
    try:
        decode_func = RECORD_DECODE_FUNCS[fmt]
        return decode_func(bvalue)
    except Exception:
        val = bytes(bvalue).decode()
        raise ServerError(
            f"Error decoding value {val}, format: {chr(fmt)}"
        )

def getitem_with_default(item: Dict, field: Any) -> Any:
    if not isinstance(item, Dict):
        raise ServerError(
            f"Cannot reduce a value of type {type(item)}")
    if field not in item:
        item[field] = {}
    return item[field]

def parse_namespace_key(key: Union[List[str], str]) -> List[str]:
    try:
        key_list = key if isinstance(key, list) else key.split('.')
    except Exception:
        key_list = []
    if not key_list or "" in key_list:
        raise ServerError(f"Invalid Key Format: '{key}'")
    return key_list

def generate_lmdb_entries(
    db_folder: pathlib.Path
) -> Generator[Tuple[str, str, bytes], Any, None]:
    if not db_folder.joinpath("data.mdb").is_file():
        return
    MAX_LMDB_NAMESPACES = 100
    MAX_LMDB_SIZE = 200 * 2**20
    inst_attempted: bool = False
    while True:
        try:
            import lmdb
            lmdb_env: LmdbEnvironment = lmdb.open(
                str(db_folder), map_size=MAX_LMDB_SIZE, max_dbs=MAX_LMDB_NAMESPACES
            )
        except ModuleNotFoundError:
            if inst_attempted:
                logging.info(
                    "Attempt to install LMDB failed, aborting conversion."
                )
                return
            import sys
            from ..utils import pip_utils
            inst_attempted = True
            logging.info("LMDB module not found, attempting installation...")
            pip_cmd = f"{sys.executable} -m pip"
            pip_exec = pip_utils.PipExecutor(pip_cmd, logging.info)
            pip_exec.install_packages(["lmdb"])
        except Exception:
            logging.exception(
                "Failed to open lmdb database, aborting conversion"
            )
            return
        else:
            break
    lmdb_namespaces: List[Tuple[str, object]] = []
    with lmdb_env.begin(buffers=True) as txn:
        # lookup existing namespaces
        with txn.cursor() as cursor:
            remaining = cursor.first()
            while remaining:
                key = bytes(cursor.key())
                if not key:
                    continue
                db = lmdb_env.open_db(key, txn)
                lmdb_namespaces.append((key.decode(), db))
                remaining = cursor.next()
        # Copy all records
        for (ns, db) in lmdb_namespaces:
            logging.info(f"Converting LMDB namespace '{ns}'")
            with txn.cursor(db=db) as cursor:
                remaining = cursor.first()
                while remaining:
                    key_buf = cursor.key()
                    value = b""
                    try:
                        decoded_key = bytes(key_buf).decode()
                        value = bytes(cursor.value())
                    except Exception:
                        logging.info("Database Key/Value Decode Error")
                        decoded_key = ''
                    remaining = cursor.next()
                    if not decoded_key or not value:
                        hk = bytes(key_buf).hex()
                        logging.info(
                            f"Invalid key or value '{hk}' found in "
                            f"lmdb namespace '{ns}'"
                        )
                        continue
                    if ns == "moonraker":
                        if decoded_key == "database":
                            # Convert "database" field in the "moonraker" namespace
                            # to its own namespace if possible
                            db_info = decode_record(value)
                            if isinstance(db_info, dict):
                                for db_key, db_val in db_info.items():
                                    yield ("database", db_key, encode_record(db_val))
                                continue
                        elif decoded_key == "database_version":
                            yield ("database", decoded_key, value)
                            continue
                    yield (ns, decoded_key, value)
    lmdb_env.close()

class MoonrakerDatabase:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.eventloop = self.server.get_event_loop()
        self.registered_namespaces: Set[str] = set(["moonraker", "database"])
        self.registered_tables: Set[str] = set([NAMESPACE_TABLE, REGISTRATION_TABLE])
        self.backup_lock = Lock()
        instance_id: str = self.server.get_app_args()["instance_uuid"]
        db_path = self._get_database_folder(config)
        self._sql_db = db_path.joinpath(SQL_DB_FILENAME)
        self.db_provider = SqliteProvider(config, self._sql_db)
        stored_iid = self.get_item("moonraker", "instance_id", None).result()
        if stored_iid is not None:
            if instance_id != stored_iid:
                self.server.add_log_rollover_item(
                    "uuid_mismatch",
                    "Database: Stored Instance ID does not match current Instance "
                    f"ID.\n\nCurrent UUID: {instance_id}\nStored UUID: {stored_iid}"
                )
        else:
            self.insert_item("moonraker", "instance_id", instance_id)
        dbinfo: Dict[str, Any] = self.get_item("database", default={}).result()
        # Protected Namespaces have read-only API access.  Write access can
        # be granted by enabling the debug option.  Forbidden namespaces
        # have no API access.  This cannot be overridden.
        ptns: Set[str] = set(dbinfo.get("protected_namespaces", []))
        fbns: Set[str] = set(dbinfo.get("forbidden_namespaces", []))
        self.protected_namespaces: Set[str] = ptns.union(["moonraker"])
        self.forbidden_namespaces: Set[str] = fbns.union(["database"])
        # Initialize Debug Counter
        config.getboolean("enable_database_debug", False, deprecate=True)
        self.debug_counter: Dict[str, int] = {"get": 0, "post": 0, "delete": 0}
        db_counter: Optional[Dict[str, int]] = dbinfo.get("debug_counter")
        if isinstance(db_counter, dict):
            self.debug_counter.update(db_counter)
            self.server.add_log_rollover_item(
                "database_debug_counter",
                f"Database Debug Counter: {self.debug_counter}"
            )
        # Track unsafe shutdowns
        self.unsafe_shutdowns: int = dbinfo.get("unsafe_shutdowns", 0)
        msg = f"Unsafe Shutdown Count: {self.unsafe_shutdowns}"
        self.server.add_log_rollover_item("database", msg)
        self.insert_item("database", "database_version", DATABASE_VERSION)
        self.server.register_endpoint(
            "/server/database/list", RequestType.GET, self._handle_list_request
        )
        self.server.register_endpoint(
            "/server/database/item", RequestType.all(), self._handle_item_request
        )
        self.server.register_endpoint(
            "/server/database/backup", RequestType.POST | RequestType.DELETE,
            self._handle_backup_request
        )
        self.server.register_endpoint(
            "/server/database/restore", RequestType.POST, self._handle_restore_request
        )
        self.server.register_endpoint(
            "/server/database/compact", RequestType.POST, self._handle_compact_request
        )
        self.server.register_debug_endpoint(
            "/debug/database/list", RequestType.GET, self._handle_list_request
        )
        self.server.register_debug_endpoint(
            "/debug/database/item", RequestType.all(), self._handle_item_request
        )
        self.server.register_debug_endpoint(
            "/debug/database/table", RequestType.GET, self._handle_table_request
        )
        # self.server.register_debug_endpoint(
        #    "/debug/database/row", RequestType.all(),
        #    self._handle_row_request
        # )

    async def component_init(self) -> None:
        await self.db_provider.async_init()
        # Increment unsafe shutdown counter.  This will be reset if moonraker is
        # safely restarted
        await self.insert_item(
            "database", "unsafe_shutdowns", self.unsafe_shutdowns + 1
        )

    def get_database_path(self) -> str:
        return str(self._sql_db)

    @property
    def database_path(self) -> pathlib.Path:
        return self._sql_db

    def _get_database_folder(self, config: ConfigHelper) -> pathlib.Path:
        app_args = self.server.get_app_args()
        dep_path = config.get("database_path", None, deprecate=True)
        db_path = pathlib.Path(app_args["data_path"]).joinpath("database")
        if (
            app_args["is_default_data_path"] and
            not db_path.joinpath(SQL_DB_FILENAME).exists()
        ):
            # Allow configured DB fallback
            dep_path = dep_path or "~/.moonraker_database"
            legacy_db = pathlib.Path(dep_path).expanduser().resolve()
            try:
                same = legacy_db.samefile(db_path)
            except Exception:
                same = False
            if not same and legacy_db.joinpath("data.mdb").is_file():
                logging.info(
                    f"Reverting to legacy database folder: {legacy_db}"
                )
                db_path = legacy_db
        if not db_path.is_dir():
            db_path.mkdir()
        return db_path

    # *** Nested Database operations***
    # The insert_item(), delete_item(), and get_item() methods may operate on
    # nested objects within a namespace.  Each operation takes a key argument
    # that may either be a string or a list of strings.  If the argument is
    # a string nested keys may be delitmted by a "." by which the string
    # will be split into a list of strings.  The first key in the list must
    # identify the database record.  Subsequent keys are optional and are
    # used to access elements in the deserialized objects.

    def insert_item(
        self, namespace: str, key: Union[List[str], str], value: DBType
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.insert_item, namespace, key, value
        )

    def update_item(
        self, namespace: str, key: Union[List[str], str], value: DBType
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.update_item, namespace, key, value
        )

    def delete_item(
        self, namespace: str, key: Union[List[str], str]
    ) -> Future[Any]:
        return self.db_provider.execute_db_function(
            self.db_provider.delete_item, namespace, key
        )

    def get_item(
        self,
        namespace: str,
        key: Optional[Union[List[str], str]] = None,
        default: Any = Sentinel.MISSING
    ) -> Future[Any]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_item, namespace, key, default
        )

    # *** Batch operations***
    #  The insert_batch(), move_batch(), delete_batch(), and get_batch()
    #  methods can be used to perform record level batch operations on
    #  a namespace in a single transaction.

    def insert_batch(
        self, namespace: str, records: Dict[str, Any]
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.insert_batch, namespace, records
        )

    def move_batch(
        self, namespace: str, source_keys: List[str], dest_keys: List[str]
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.move_batch, namespace, source_keys, dest_keys
        )

    def delete_batch(
        self, namespace: str, keys: List[str]
    ) -> Future[Dict[str, Any]]:
        return self.db_provider.execute_db_function(
            self.db_provider.delete_batch, namespace, keys
        )

    def get_batch(
        self, namespace: str, keys: List[str]
    ) -> Future[Dict[str, Any]]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_batch, namespace, keys
        )

    # *** Namespace level operations***

    def update_namespace(
        self, namespace: str, values: Dict[str, DBRecord]
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.insert_batch, namespace, values
        )

    def clear_namespace(self, namespace: str) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.clear_namespace, namespace
        )

    def sync_namespace(
        self, namespace: str, values: Dict[str, DBRecord]
    ) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.sync_namespace, namespace, values
        )

    def ns_length(self, namespace: str) -> Future[int]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_namespace_length, namespace
        )

    def ns_keys(self, namespace: str) -> Future[List[str]]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_namespace_keys, namespace,
        )

    def ns_values(self, namespace: str) -> Future[List[Any]]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_namespace_values, namespace
        )

    def ns_items(self, namespace: str) -> Future[List[Tuple[str, Any]]]:
        return self.db_provider.execute_db_function(
            self.db_provider.get_namespace_items, namespace
        )

    def ns_contains(
        self, namespace: str, key: Union[List[str], str]
    ) -> Future[bool]:
        return self.db_provider.execute_db_function(
            self.db_provider.namespace_contains, namespace
        )

    # SQL direct query methods
    def sql_execute(
        self, sql: str, params: SqlParams = []
    ) -> Future[SqliteCursorProxy]:
        return self.db_provider.execute_db_function(
            self.db_provider.sql_execute, sql, params
        )

    def sql_executemany(
        self, sql: str, params: Sequence[SqlParams] = []
    ) -> Future[SqliteCursorProxy]:
        return self.db_provider.execute_db_function(
            self.db_provider.sql_executemany, sql, params
        )

    def sql_executescript(self, sql: str) -> Future[SqliteCursorProxy]:
        return self.db_provider.execute_db_function(
            self.db_provider.sql_executescript, sql
        )

    def sql_commit(self) -> Future[None]:
        return self.db_provider.execute_db_function(self.db_provider.sql_commit)

    def sql_rollback(self) -> Future[None]:
        return self.db_provider.execute_db_function(self.db_provider.sql_rollback)

    def queue_sql_callback(
        self, callback: Callable[[sqlite3.Connection], Any]
    ) -> Future[Any]:
        return self.db_provider.execute_db_function(callback)

    def compact_database(self) -> Future[Dict[str, int]]:
        return self.db_provider.execute_db_function(
            self.db_provider.compact_database
        )

    def backup_database(self, bkp_path: pathlib.Path) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.backup_database, bkp_path
        )

    def restore_database(self, restore_path: pathlib.Path) -> Future[Dict[str, Any]]:
        return self.db_provider.execute_db_function(
            self.db_provider.restore_database, restore_path
        )

    def register_local_namespace(
            self, namespace: str, forbidden: bool = False, parse_keys: bool = False
    ) -> NamespaceWrapper:
        if namespace in self.registered_namespaces:
            raise self.server.error(f"Namespace '{namespace}' already registered")
        self.registered_namespaces.add(namespace)
        self.db_provider.register_namespace(namespace)
        if forbidden:
            if namespace not in self.forbidden_namespaces:
                self.forbidden_namespaces.add(namespace)
                self.insert_item(
                    "database", "forbidden_namespaces",
                    sorted(self.forbidden_namespaces)
                )
        elif namespace not in self.protected_namespaces:
            self.protected_namespaces.add(namespace)
            self.insert_item(
                "database", "protected_namespaces", sorted(self.protected_namespaces)
            )
        return NamespaceWrapper(namespace, self, parse_keys)

    def wrap_namespace(
        self, namespace: str, parse_keys: bool = True
    ) -> NamespaceWrapper:
        if namespace not in self.db_provider.namespaces:
            raise self.server.error(f"Namespace '{namespace}' not found", 404)
        return NamespaceWrapper(namespace, self, parse_keys)

    def unregister_local_namespace(self, namespace: str) -> None:
        if namespace in self.registered_namespaces:
            self.registered_namespaces.remove(namespace)
        if namespace in self.forbidden_namespaces:
            self.forbidden_namespaces.remove(namespace)
            self.insert_item(
                "database", "forbidden_namespaces", sorted(self.forbidden_namespaces)
            )
        if namespace in self.protected_namespaces:
            self.protected_namespaces.remove(namespace)
            self.insert_item(
                "database", "protected_namespaces", sorted(self.protected_namespaces)
            )

    def drop_empty_namespace(self, namespace: str) -> Future[None]:
        return self.db_provider.execute_db_function(
            self.db_provider.drop_empty_namespace, namespace
        )

    def get_provider_wrapper(self) -> DBProviderWrapper:
        return self.db_provider.get_provider_wapper()

    def get_backup_dir(self) -> pathlib.Path:
        bkp_dir = pathlib.Path(self.server.get_app_arg("data_path"))
        return bkp_dir.joinpath("backup/database").resolve()

    def register_table(self, table_def: SqlTableDefinition) -> SqlTableWrapper:
        if table_def.name in self.registered_tables:
            raise self.server.error(f"Table '{table_def.name}' already registered")
        self.registered_tables.add(table_def.name)
        self.db_provider.register_table(table_def)
        return SqlTableWrapper(self, table_def)

    async def _handle_compact_request(self, web_request: WebRequest) -> Dict[str, int]:
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        if kconn.is_printing():
            raise self.server.error("Cannot compact when Klipper is printing")
        async with self.backup_lock:
            return await self.compact_database()

    async def _handle_backup_request(self, web_request: WebRequest) -> Dict[str, Any]:
        async with self.backup_lock:
            request_type = web_request.get_request_type()
            if request_type == RequestType.POST:
                kconn: KlippyConnection
                kconn = self.server.lookup_component("klippy_connection")
                if kconn.is_printing():
                    raise self.server.error("Cannot backup when Klipper is printing")
                suffix = time.strftime("%Y%m%d-%H%M%S", time.localtime())
                db_name = web_request.get_str("filename", f"sqldb-backup-{suffix}.db")
                bkp_dir = self.get_backup_dir()
                bkp_path = bkp_dir.joinpath(db_name).resolve()
                if bkp_dir not in bkp_path.parents:
                    raise self.server.error(f"Invalid name {db_name}.")
                await self.backup_database(bkp_path)
            elif request_type == RequestType.DELETE:
                db_name = web_request.get_str("filename")
                bkp_dir = self.get_backup_dir()
                bkp_path = bkp_dir.joinpath(db_name).resolve()
                if bkp_dir not in bkp_path.parents:
                    raise self.server.error(f"Invalid name {db_name}.")
                if not bkp_path.is_file():
                    raise self.server.error(
                        f"Backup file {db_name} does not exist", 404
                    )
                await self.eventloop.run_in_thread(bkp_path.unlink)
            else:
                raise self.server.error("Invalid request type")
            return {
                "backup_path": str(bkp_path)
            }

    async def _handle_restore_request(self, web_request: WebRequest) -> Dict[str, Any]:
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        if kconn.is_printing():
            raise self.server.error("Cannot restore when Klipper is printing")
        async with self.backup_lock:
            db_name = web_request.get_str("filename")
            bkp_dir = self.get_backup_dir()
            restore_path = bkp_dir.joinpath(db_name).resolve()
            if bkp_dir not in restore_path.parents:
                raise self.server.error(f"Invalid name {db_name}.")
            restore_info = await self.restore_database(restore_path)
            self.server.restart(.1)
            return restore_info

    async def _handle_list_request(
        self, web_request: WebRequest
    ) -> Dict[str, List[str]]:
        path = web_request.get_endpoint()
        ns_list = set(self.db_provider.namespaces)
        bkp_dir = self.get_backup_dir()
        backups: List[str] = []
        if bkp_dir.is_dir():
            backups = [bkp.name for bkp in bkp_dir.iterdir() if bkp.is_file()]
        if not path.startswith("/debug/"):
            ns_list -= self.forbidden_namespaces
            return {
                "namespaces": list(ns_list),
                "backups": backups
            }
        else:
            return {
                "namespaces": list(ns_list),
                "backups": backups,
                "tables": list(self.db_provider.tables)
            }

    async def _handle_item_request(self, web_request: WebRequest) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        is_debug = web_request.get_endpoint().startswith("/debug/")
        namespace = web_request.get_str("namespace")
        if namespace in self.forbidden_namespaces and not is_debug:
            raise self.server.error(
                f"Read/Write access to namespace '{namespace}' is forbidden", 403
            )
        if req_type == RequestType.GET:
            key = web_request.get("key", None)
            if key is not None and not isinstance(key, (list, str)):
                raise self.server.error(
                    "Value for argument 'key' is an invalid type: "
                    f"{type(key).__name__}"
                )
            val = await self.get_item(namespace, key)
        else:
            if namespace in self.protected_namespaces and not is_debug:
                raise self.server.error(
                    f"Write access to namespace '{namespace}' is forbidden", 403
                )
            key = web_request.get("key")
            if not isinstance(key, (list, str)):
                raise self.server.error(
                    "Value for argument 'key' is an invalid type: "
                    f"{type(key).__name__}"
                )
            if req_type == RequestType.POST:
                val = web_request.get("value")
                await self.insert_item(namespace, key, val)
            elif req_type == RequestType.DELETE:
                val = await self.delete_item(namespace, key)
                await self.drop_empty_namespace(namespace)
            else:
                raise self.server.error(f"Invalid request type {req_type}")

        if is_debug:
            name = req_type.name or str(req_type).split(".", 1)[-1]
            self.debug_counter[name.lower()] += 1
            await self.insert_item(
                "database", "debug_counter", self.debug_counter
            )
            self.server.add_log_rollover_item(
                "database_debug_counter",
                f"Database Debug Counter: {self.debug_counter}",
                log=False
            )
        return {'namespace': namespace, 'key': key, 'value': val}

    async def close(self) -> None:
        if not self.db_provider.is_restored():
            # Don't overwrite unsafe shutdowns on a restored database
            await self.insert_item(
                "database", "unsafe_shutdowns", self.unsafe_shutdowns
            )
        # Stop command thread
        await self.db_provider.stop()

    async def _handle_table_request(self, web_request: WebRequest) -> Dict[str, Any]:
        table = web_request.get_str("table")
        if table not in self.db_provider.tables:
            raise self.server.error(f"Table name '{table}' does not exist", 404)
        cur = await self.sql_execute(f"SELECT rowid, * FROM {table}")
        return {
            "table_name": table,
            "rows": [dict(r) for r in await cur.fetchall()]
        }

    async def _handle_row_request(self, web_request: WebRequest) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        table = web_request.get_str("table")
        if table not in self.db_provider.tables:
            raise self.server.error(
                f"Table name '{table}' does not exist", 404
            )
        if req_type == RequestType.POST:
            row_id = web_request.get_int("id", None)
            values = web_request.get("values")
            assert isinstance(values, dict)
            keys = set(values.keys())
            cur = await self.sql_execute(f"PRAGMA table_info('{table}')")
            columns = set([r["name"] for r in await cur.fetchall()])
            if row_id is None:
                # insert
                if keys != columns:
                    raise self.server.error(
                        "Keys in value to insert do not match columns of tables"
                    )
                val_str = ",".join([f":{col}" for col in columns])
                cur = await self.sql_execute(
                    f"INSERT INTO {table} VALUES({val_str})", values
                )
            else:
                # update
                if not keys.issubset(columns):
                    raise self.server.error(
                        "Keys in value to update are not a subset of available columns"
                    )
                col_str = ",".join([f"{col}" for col in columns if col in keys])
                vals = [values[col] for col in columns if col in keys]
                vals.append(row_id)
                val_str = ",".join("?" * len(vals))
                cur = await self.sql_execute(
                    f"UPDATE {table} SET ({col_str}) = ({val_str}) WHERE rowid = ?",
                    vals
                )
                if not cur.rowcount:
                    raise self.server.error(f"No row with id {row_id} to update")
        else:
            row_id = web_request.get_int("id")
        cur = await self.sql_execute(
            f"SELECT rowid, * FROM {table} WHERE rowid = ?", (row_id,)
        )
        item = dict(await cur.fetchone() or {})
        if req_type == RequestType.DELETE:
            await self.sql_execute(
                f"DELETE FROM {table} WHERE rowid = ?", (row_id,)
            )
        return {
            "row": item
        }

class SqliteProvider(Thread):
    def __init__(self, config: ConfigHelper, db_path: pathlib.Path) -> None:
        super().__init__()
        self.server = config.get_server()
        self.asyncio_loop = self.server.get_event_loop().asyncio_loop
        self._namespaces: Set[str] = set()
        self._tables: Set[str] = set()
        self._db_path = db_path
        self.restored: bool = False
        self.command_queue: Queue[Tuple[Future, Optional[Callable], Tuple[Any, ...]]]
        self.command_queue = Queue()
        sqlite3.register_converter("record", decode_record)
        sqlite3.register_converter("pyjson", jsonw.loads)
        sqlite3.register_converter("pybool", lambda x: bool(x))
        sqlite3.register_adapter(list, jsonw.dumps)
        sqlite3.register_adapter(dict, jsonw.dumps)
        self.sync_conn = sqlite3.connect(
            str(db_path), timeout=1., detect_types=sqlite3.PARSE_DECLTYPES
        )
        self.sync_conn.row_factory = sqlite3.Row
        self.setup_database()

    @property
    def namespaces(self) -> Set[str]:
        return self._namespaces

    @property
    def tables(self) -> Set[str]:
        return self._tables

    def async_init(self) -> Future[str]:
        self.sync_conn.close()
        self.start()
        fut = self.asyncio_loop.create_future()
        self.command_queue.put_nowait((fut, lambda x: "sqlite", tuple()))
        return fut

    def run(self) -> None:
        loop = self.asyncio_loop
        conn = sqlite3.connect(
            str(self._db_path), timeout=1., detect_types=sqlite3.PARSE_DECLTYPES
        )
        conn.row_factory = sqlite3.Row
        while True:
            future, func, args = self.command_queue.get()
            if func is None:
                break
            try:
                ret = func(conn, *args)
            except Exception as e:
                loop.call_soon_threadsafe(future.set_exception, e)
            else:
                loop.call_soon_threadsafe(future.set_result, ret)
        conn.close()
        loop.call_soon_threadsafe(future.set_result, None)

    def execute_db_function(
        self, command_func: Callable[..., _T], *args
    ) -> Future[_T]:
        fut = self.asyncio_loop.create_future()
        if self.is_alive():
            self.command_queue.put_nowait((fut, command_func, args))
        else:
            ret = command_func(self.sync_conn, *args)
            fut.set_result(ret)
        return fut

    def setup_database(self) -> None:
        self.server.add_log_rollover_item(
            "sqlite_intro",
            "Loading Sqlite database provider. "
            f"Sqlite Version: {sqlite3.sqlite_version}"
        )
        cur = self.sync_conn.execute(
            f"SELECT name FROM {SCHEMA_TABLE} WHERE type='table'"
        )
        cur.arraysize = 100
        self._tables = set([row[0] for row in cur.fetchall()])
        logging.debug(f"Detected SQL Tables: {self._tables}")
        if NAMESPACE_TABLE not in self._tables:
            self._create_default_tables()
            self._migrate_from_lmdb()
        elif REGISTRATION_TABLE not in self._tables:
            self._create_registration_table()
        # Find namespaces
        cur = self.sync_conn.execute(
            f"SELECT DISTINCT namespace FROM {NAMESPACE_TABLE}"
        )
        cur.arraysize = 100
        self._namespaces = set([row[0] for row in cur.fetchall()])
        logging.debug(f"Detected namespaces: {self._namespaces}")

    def _migrate_from_lmdb(self) -> None:
        db_folder = self._db_path.parent
        if not db_folder.joinpath("data.mdb").is_file():
            return
        logging.info("Converting LMDB Database to Sqlite...")
        with self.sync_conn:
            self.sync_conn.executemany(
                f"INSERT INTO {NAMESPACE_TABLE} VALUES (?,?,?)",
                generate_lmdb_entries(db_folder)
            )

    def _create_default_tables(self) -> None:
        self._create_registration_table()
        if NAMESPACE_TABLE in self._tables:
            return
        namespace_proto = inspect.cleandoc(
            f"""
            {NAMESPACE_TABLE} (
                namespace TEXT NOT NULL,
                key TEXT NOT NULL,
                value record NOT NULL,
                PRIMARY KEY (namespace, key)
            )
            """
        )
        with self.sync_conn:
            self.sync_conn.execute(f"CREATE TABLE {namespace_proto}")
        self._save_registered_table(NAMESPACE_TABLE, namespace_proto, 1)
        self.server.add_log_rollover_item(
            "db_default_table", f"Created default SQL table {NAMESPACE_TABLE}"
        )

    def _create_registration_table(self) -> None:
        if REGISTRATION_TABLE in self._tables:
            return
        reg_tbl_proto = inspect.cleandoc(
            f"""
            {REGISTRATION_TABLE} (
                name TEXT NOT NULL PRIMARY KEY,
                prototype TEXT NOT NULL,
                version INT
            )
            """
        )
        with self.sync_conn:
            self.sync_conn.execute(f"CREATE TABLE {reg_tbl_proto}")
        self._tables.add(REGISTRATION_TABLE)

    def _save_registered_table(
        self, table_name: str, prototype: str, version: int
    ) -> None:
        with self.sync_conn:
            self.sync_conn.execute(
                f"INSERT INTO {REGISTRATION_TABLE} VALUES(?, ?, ?) "
                "ON CONFLICT(name) DO UPDATE SET "
                "prototype=excluded.prototype, version=excluded.version",
                (table_name, prototype, version)
            )
        self._tables.add(table_name)

    def _lookup_registered_table(self, table_name: str) -> Tuple[str, int]:
        cur = self.sync_conn.execute(
            f"SELECT prototype, version FROM {REGISTRATION_TABLE} "
            f"WHERE name = ?",
            (table_name,)
        )
        ret = cur.fetchall()
        if not ret:
            return "", 0
        return tuple(ret[0])  # type: ignore

    def _insert_record(
        self, conn: sqlite3.Connection, namespace: str, key: str, val: DBType
    ) -> bool:
        if val is None:
            return False
        try:
            with conn:
                conn.execute(
                    f"INSERT INTO {NAMESPACE_TABLE} VALUES(?, ?, ?) "
                    "ON CONFLICT(namespace, key) DO UPDATE SET value=excluded.value",
                    (namespace, key, encode_record(val))
                )
        except sqlite3.Error:
            if self.server.is_verbose_enabled():
                logging.error("Error inserting record for key")
            return False
        return True

    def _get_record(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        key: str,
        default: Union[Sentinel, DBRecord] = Sentinel.MISSING
    ) -> DBRecord:
        cur = conn.execute(
            f"SELECT value FROM {NAMESPACE_TABLE} WHERE namespace = ? and key = ?",
            (namespace, key)
        )
        val = cur.fetchone()
        if val is None:
            if default is Sentinel.MISSING:
                raise self.server.error(
                    f"Key '{key}' in namespace '{namespace}' not found", 404
                )
            return default
        return val[0]

    # Namespace Query Ops

    def get_namespace(
        self, conn: sqlite3.Connection, namespace: str, must_exist: bool = True
    ) -> Dict[str, Any]:
        if namespace not in self._namespaces:
            if not must_exist:
                return {}
            raise self.server.error(f"Namespace {namespace} not found", 404)
        cur = conn.execute(
            f"SELECT key, value FROM {NAMESPACE_TABLE} WHERE namespace = ?",
            (namespace,)
        )
        cur.arraysize = 200
        return dict(cur.fetchall())

    def iter_namespace(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        count: int = 1000
    ) -> Generator[Dict[str, Any], Any, None]:
        if self.is_alive():
            raise self.server.error("Cannot iterate a namespace asynchronously")
        if namespace not in self._namespaces:
            return
        offset: int = 0
        total = self.get_namespace_length(conn, namespace)
        while offset < total:
            cur = conn.execute(
                f"SELECT key, value FROM {NAMESPACE_TABLE} WHERE namespace = ? "
                f"LIMIT ? OFFSET ?",
                (namespace, count, offset)
            )
            cur.arraysize = count
            ret = cur.fetchall()
            if not ret:
                return
            yield dict(ret)
            offset += count

    def clear_namespace(self, conn: sqlite3.Connection, namespace: str) -> None:
        with conn:
            conn.execute(
                f"DELETE FROM {NAMESPACE_TABLE} WHERE namespace = ?", (namespace,)
            )

    def drop_empty_namespace(self, conn: sqlite3.Connection, namespace: str) -> None:
        if namespace in self._namespaces:
            if self.get_namespace_length(conn, namespace) == 0:
                self._namespaces.remove(namespace)

    def sync_namespace(
        self, conn: sqlite3.Connection, namespace: str, values: Dict[str, DBRecord]
    ) -> None:
        def generate_params():
            for key, val in values.items():
                yield (namespace, key, val)
        with conn:
            conn.execute(
                f"DELETE FROM {NAMESPACE_TABLE} WHERE namespace = ?", (namespace,)
            )
            conn.executemany(
                f"INSERT INTO {NAMESPACE_TABLE} VALUES(?, ?, ?)", generate_params()
            )

    def get_namespace_length(self, conn: sqlite3.Connection, namespace: str) -> int:
        cur = conn.execute(
            f"SELECT COUNT(namespace) FROM {NAMESPACE_TABLE} WHERE namespace = ?",
            (namespace,)
        )
        return cur.fetchone()[0]

    def get_namespace_keys(self, conn: sqlite3.Connection, namespace: str) -> List[str]:
        cur = conn.execute(
            f"SELECT key FROM {NAMESPACE_TABLE} WHERE namespace = ?",
            (namespace,)
        )
        cur.arraysize = 200
        return [row[0] for row in cur.fetchall()]

    def get_namespace_values(
        self, conn: sqlite3.Connection, namespace: str
    ) -> List[Any]:
        cur = conn.execute(
            f"SELECT value FROM {NAMESPACE_TABLE} WHERE namespace = ?",
            (namespace,)
        )
        cur.arraysize = 200
        return [row[0] for row in cur.fetchall()]

    def get_namespace_items(
        self, conn: sqlite3.Connection, namespace: str
    ) -> List[Tuple[str, Any]]:
        cur = conn.execute(
            f"SELECT key, value FROM {NAMESPACE_TABLE} WHERE namespace = ?",
            (namespace,)
        )
        cur.arraysize = 200
        return cur.fetchall()

    def namespace_contains(
        self, conn: sqlite3.Connection, namespace: str, key: Union[List[str], str]
    ) -> bool:
        try:
            key_list = parse_namespace_key(key)
            if len(key_list) == 1:
                cur = conn.execute(
                    f"SELECT key FROM {NAMESPACE_TABLE} "
                    "WHERE namespace = ? and key = ?",
                    (namespace, key)
                )
                return cur.fetchone() is not None
            record = self._get_record(conn, namespace, key_list[0])
            reduce(operator.getitem, key_list[1:], record)  # type: ignore
        except Exception:
            return False
        return True

    def insert_item(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        key: Union[List[str], str],
        value: DBType
    ) -> None:
        key_list = parse_namespace_key(key)
        record = value
        if len(key_list) > 1:
            record = self._get_record(conn, namespace, key_list[0], default={})
            if not isinstance(record, dict):
                prev_type = type(record)
                record = {}
                logging.info(
                    f"Warning: Key {key_list[0]} contains a value of type "
                    f"{prev_type}. Overwriting with an object."
                )
            item: DBType = reduce(getitem_with_default, key_list[1:-1], record)
            if not isinstance(item, dict):
                rpt_key = ".".join(key_list[:-1])
                raise self.server.error(
                    f"Item at key '{rpt_key}' in namespace '{namespace}'is "
                    "not a dictionary object, cannot insert"
                )
            item[key_list[-1]] = value
        if not self._insert_record(conn, namespace, key_list[0], record):
            logging.info(f"Error inserting key '{key}' in namespace '{namespace}'")
        else:
            self._namespaces.add(namespace)

    def update_item(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        key: Union[List[str], str],
        value: DBType
    ) -> None:
        key_list = parse_namespace_key(key)
        record = self._get_record(conn, namespace, key_list[0])
        if len(key_list) == 1:
            if isinstance(record, dict) and isinstance(value, dict):
                record.update(value)
            else:
                record = value
        else:
            try:
                assert isinstance(record, dict)
                item: Dict[str, Any] = reduce(
                    operator.getitem, key_list[1:-1], record
                )
            except Exception:
                raise self.server.error(
                    f"Key '{key}' in namespace '{namespace}' not found", 404
                )
            if not isinstance(item, dict) or key_list[-1] not in item:
                rpt_key = ".".join(key_list[:-1])
                raise self.server.error(
                    f"Item at key '{rpt_key}' in namespace '{namespace}'is "
                    "not a dictionary object, cannot update"
                )
            if isinstance(item[key_list[-1]], dict) and isinstance(value, dict):
                item[key_list[-1]].update(value)
            else:
                item[key_list[-1]] = value
        if not self._insert_record(conn, namespace, key_list[0], record):
            logging.info(f"Error updating key '{key}' in namespace '{namespace}'")

    def delete_item(
        self, conn: sqlite3.Connection, namespace: str, key: Union[List[str], str]
    ) -> Any:
        key_list = parse_namespace_key(key)
        val = record = self._get_record(conn, namespace, key_list[0])
        remove_record = True
        if len(key_list) > 1:
            try:
                assert isinstance(record, dict)
                item: Dict[str, Any] = reduce(
                    operator.getitem, key_list[1:-1], record)
                val = item.pop(key_list[-1])
            except Exception:
                raise self.server.error(
                    f"Key '{key}' in namespace '{namespace}' not found",
                    404)
            remove_record = False if record else True
        if remove_record:
            with conn:
                conn.execute(
                    f"DELETE FROM {NAMESPACE_TABLE} WHERE namespace = ? and key = ?",
                    (namespace, key_list[0])
                )
        else:
            ret = self._insert_record(conn, namespace, key_list[0], record)
            if not ret:
                logging.info(
                    f"Error deleting key '{key}' from namespace '{namespace}'"
                )
        return val

    def get_item(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        key: Optional[Union[List[str], str]] = None,
        default: Any = Sentinel.MISSING
    ) -> Any:
        try:
            if key is None:
                return self.get_namespace(conn, namespace)
            key_list = parse_namespace_key(key)
            rec = self._get_record(conn, namespace, key_list[0])
            val = reduce(operator.getitem, key_list[1:], rec)  # type: ignore
        except Exception as e:
            if default is not Sentinel.MISSING:
                return default
            if isinstance(e, self.server.error):
                raise
            raise self.server.error(
                f"Key '{key}' in namespace '{namespace}' not found", 404
            )
        return val

    def insert_batch(
        self, conn: sqlite3.Connection, namespace: str, records: Dict[str, Any]
    ) -> None:
        def generate_params():
            for key, val in records.items():
                yield (namespace, key, encode_record(val))
        with conn:
            conn.executemany(
                f"INSERT INTO {NAMESPACE_TABLE} VALUES(?, ?, ?) "
                "ON CONFLICT(namespace, key) DO UPDATE SET value=excluded.value",
                generate_params()
            )
        self._namespaces.add(namespace)

    def move_batch(
        self,
        conn: sqlite3.Connection,
        namespace: str,
        source_keys: List[str],
        dest_keys: List[str]
    ) -> None:
        def generate_params():
            for src, dest in zip(source_keys, dest_keys):
                yield (dest, namespace, src)
        with conn:
            conn.executemany(
                f"UPDATE OR REPLACE {NAMESPACE_TABLE} SET key = ? "
                "WHERE namespace = ? and key = ?",
                generate_params()
            )

    def delete_batch(
        self, conn: sqlite3.Connection, namespace: str, keys: List[str]
    ) -> Dict[str, Any]:
        def generate_params():
            for key in keys:
                yield (namespace, key)
        if sqlite3.sqlite_version_info < (3, 35):
            vals = self.get_batch(conn, namespace, keys)
            with conn:
                conn.executemany(
                    f"DELETE FROM {NAMESPACE_TABLE} WHERE namespace = ? and key = ?",
                    generate_params()
                )
            return vals
        else:
            placeholders = ",".join("?" * len(keys))
            sql = (
                f"DELETE FROM {NAMESPACE_TABLE} "
                f"WHERE namespace = ? and key IN ({placeholders}) "
                "RETURNING key, value"
            )
            params = [namespace] + keys
            with conn:
                cur = conn.execute(sql, params)
                cur.arraysize = 200
                return dict(cur.fetchall())

    def get_batch(
        self, conn: sqlite3.Connection, namespace: str, keys: List[str]
    ) -> Dict[str, Any]:
        placeholders = ",".join("?" * len(keys))
        sql = (
            f"SELECT key, value FROM {NAMESPACE_TABLE} "
            f"WHERE namespace = ? and key IN ({placeholders})"
        )
        ph_vals = [namespace] + keys
        cur = conn.execute(sql, ph_vals)
        cur.arraysize = 200
        return dict(cur.fetchall())

    # SQL Direct Manipulation
    def sql_execute(
        self,
        conn: sqlite3.Connection,
        statement: str,
        params: SqlParams
    ) -> SqliteCursorProxy:
        cur = conn.execute(statement, params)
        cur.arraysize = 100
        return SqliteCursorProxy(self, cur)

    def sql_executemany(
        self,
        conn: sqlite3.Connection,
        statement: str,
        params: Sequence[SqlParams]
    ) -> SqliteCursorProxy:
        cur = conn.executemany(statement, params)
        cur.arraysize = 100
        return SqliteCursorProxy(self, cur)

    def sql_executescript(
        self,
        conn: sqlite3.Connection,
        script: str
    ) -> SqliteCursorProxy:
        cur = conn.executescript(script)
        cur.arraysize = 100
        return SqliteCursorProxy(self, cur)

    def sql_commit(self, conn: sqlite3.Connection) -> None:
        conn.commit()

    def sql_rollback(self, conn: sqlite3.Connection) -> None:
        conn.rollback()

    def register_namespace(self, namespace: str) -> None:
        self._namespaces.add(namespace)

    def register_table(self, table_def: SqlTableDefinition) -> None:
        if self.is_alive():
            raise self.server.error(
                "Table registration must occur during during init."
            )
        if table_def.name in self._tables:
            logging.info(f"Found registered table {table_def.name}")
            if table_def.name in (NAMESPACE_TABLE, REGISTRATION_TABLE):
                raise self.server.error(
                    f"Cannot register table '{table_def.name}', it is reserved"
                )
            detected_proto, version = self._lookup_registered_table(table_def.name)
        else:
            logging.info(f"Creating table {table_def.name}...")
            with self.sync_conn:
                self.sync_conn.execute(f"CREATE TABLE {table_def.prototype}")
            detected_proto = table_def.prototype
            version = 0
        if table_def.version > version:
            table_def.migrate(version, self.get_provider_wapper())
            self._save_registered_table(
                table_def.name, table_def.prototype, table_def.version
            )
        elif detected_proto != table_def.prototype:
            self.server.add_warning(
                f"Table '{table_def.name}' defintion does not match stored "
                "definition.  See the log for details."
            )
            logging.info(
                f"Expected table prototype:\n{table_def.prototype}\n\n"
                f"Stored table prototype:\n{detected_proto}"
            )

    def compact_database(self, conn: sqlite3.Connection) -> Dict[str, int]:
        if self.restored:
            raise self.server.error(
                "Cannot compact restored database, awaiting restart"
            )
        cur_size = self._db_path.stat().st_size
        conn.execute("VACUUM")
        new_size = self._db_path.stat().st_size
        return {
            "previous_size": cur_size,
            "new_size": new_size
        }

    def backup_database(
        self, conn: sqlite3.Connection, bkp_path: pathlib.Path
    ) -> None:
        if self.restored:
            raise self.server.error(
                "Cannot backup restored database, awaiting restart"
            )
        parent = bkp_path.parent
        if not parent.exists():
            parent.mkdir(parents=True, exist_ok=True)
        elif bkp_path.exists():
            bkp_path.unlink()
        bkp_conn = sqlite3.connect(str(bkp_path))
        conn.backup(bkp_conn)
        bkp_conn.close()

    def restore_database(
        self, conn: sqlite3.Connection, restore_path: pathlib.Path
    ) -> Dict[str, Any]:
        if self.restored:
            raise self.server.error("Database already restored")
        if not restore_path.is_file():
            raise self.server.error(f"Restoration File {restore_path} does not exist")
        restore_conn = sqlite3.connect(str(restore_path))
        restore_info = self._validate_restore_db(restore_conn)
        restore_conn.backup(conn)
        restore_conn.close()
        self.restored = True
        return restore_info

    def _validate_restore_db(
        self, restore_conn: sqlite3.Connection
    ) -> Dict[str, Any]:
        cursor = restore_conn.execute(
            f"SELECT name FROM {SCHEMA_TABLE} WHERE type = 'table'"
        )
        cursor.arraysize = 100
        tables = [row[0] for row in cursor.fetchall()]
        if NAMESPACE_TABLE not in tables:
            restore_conn.close()
            raise self.server.error(
                f"Invalid database for restoration, missing table '{NAMESPACE_TABLE}'"
            )
        missing_tables = self._tables.difference(tables)
        if missing_tables:
            logging.info(f"Database to restore missing tables: {missing_tables}")
        cursor = restore_conn.execute(
            f"SELECT DISTINCT namespace FROM {NAMESPACE_TABLE}"
        )
        cursor.arraysize = 100
        namespaces = [row[0] for row in cursor.fetchall()]
        missing_ns = self._namespaces.difference(namespaces)
        if missing_ns:
            logging.info(f"Database to restore missing namespaces: {missing_ns}")
        return {
            "restored_tables": tables,
            "restored_namespaces": namespaces
        }

    def get_provider_wapper(self) -> DBProviderWrapper:
        return DBProviderWrapper(self)

    def is_restored(self) -> bool:
        return self.restored

    def stop(self) -> Future[None]:
        fut = self.asyncio_loop.create_future()
        if not self.is_alive():
            fut.set_result(None)
        else:
            self.command_queue.put_nowait((fut, None, tuple()))
        return fut

class DBProviderWrapper:
    def __init__(self, provider: SqliteProvider) -> None:
        self.server = provider.server
        self.provider = provider
        self._sql_conn = provider.sync_conn

    @property
    def connection(self) -> sqlite3.Connection:
        return self._sql_conn

    def iter_namespace(
        self, namespace: str, batch_count: int = 100
    ) -> Generator[Dict[str, Any], Any, None]:
        yield from self.provider.iter_namespace(self._sql_conn, namespace, batch_count)

    def get_namespace_keys(self, namespace: str) -> List[str]:
        return self.provider.get_namespace_keys(self._sql_conn, namespace)

    def get_namespace_values(self, namespace: str) -> List[Any]:
        return self.provider.get_namespace_values(self._sql_conn, namespace)

    def get_namespace_items(self, namespace: str) -> List[Tuple[str, Any]]:
        return self.provider.get_namespace_items(self._sql_conn, namespace)

    def get_namespace_length(self, namespace: str) -> int:
        return self.provider.get_namespace_length(self._sql_conn, namespace)

    def get_namespace(self, namespace: str) -> Dict[str, Any]:
        return self.provider.get_namespace(self._sql_conn, namespace, must_exist=False)

    def clear_namespace(self, namespace: str) -> None:
        self.provider.clear_namespace(self._sql_conn, namespace)

    def get_item(
        self,
        namespace: str,
        key: Union[str, List[str]],
        default: Any = Sentinel.MISSING
    ) -> Any:
        return self.provider.get_item(self._sql_conn, namespace, key, default)

    def delete_item(self, namespace: str, key: Union[str, List[str]]) -> Any:
        return self.provider.delete_item(self._sql_conn, namespace, key)

    def insert_item(
        self, namespace: str, key: Union[str, List[str]], value: DBType
    ) -> None:
        self.provider.insert_item(self._sql_conn, namespace, key, value)

    def update_item(
        self, namespace: str, key: Union[str, List[str]], value: DBType
    ) -> None:
        self.provider.update_item(self._sql_conn, namespace, key, value)

    def get_batch(self, namespace: str, keys: List[str]) -> Dict[str, Any]:
        return self.provider.get_batch(self._sql_conn, namespace, keys)

    def delete_batch(self, namespace: str, keys: List[str]) -> Dict[str, Any]:
        return self.provider.delete_batch(self._sql_conn, namespace, keys)

    def insert_batch(self, namespace: str, records: Dict[str, Any]) -> None:
        self.provider.insert_batch(self._sql_conn, namespace, records)

    def move_batch(
        self, namespace: str, source_keys: List[str], dest_keys: List[str]
    ) -> None:
        self.provider.move_batch(self._sql_conn, namespace, source_keys, dest_keys)

    def wipe_local_namespace(self, namespace: str) -> None:
        """
        Unregister persistent local namespace
        """
        self.provider.clear_namespace(self._sql_conn, namespace)
        self.provider.drop_empty_namespace(self._sql_conn, namespace)
        db: MoonrakerDatabase = self.server.lookup_component("database")
        db.unregister_local_namespace(namespace)


class SqliteCursorProxy:
    def __init__(self, provider: SqliteProvider, cursor: sqlite3.Cursor) -> None:
        self._db_provider = provider
        self._cursor = cursor
        self._description = cursor.description
        self._rowcount = cursor.rowcount
        self._lastrowid = cursor.lastrowid
        self._array_size = cursor.arraysize

    @property
    def rowcount(self) -> int:
        return self._rowcount

    @property
    def lastrowid(self) -> Optional[int]:
        return self._lastrowid

    @property
    def description(self):
        return self._description

    @property
    def arraysize(self) -> int:
        return self._array_size

    def set_arraysize(self, size: int) -> Future[None]:
        def wrapper(_) -> None:
            self._cursor.arraysize = size
            self._array_size = size
        return self._db_provider.execute_db_function(wrapper)

    def fetchone(self) -> Future[Optional[sqlite3.Row]]:
        def fetch_wrapper(_) -> Optional[sqlite3.Row]:
            return self._cursor.fetchone()
        return self._db_provider.execute_db_function(fetch_wrapper)

    def fetchmany(self, size: Optional[int] = None) -> Future[List[sqlite3.Row]]:
        def fetch_wrapper(_) -> List[sqlite3.Row]:
            if size is None:
                return self._cursor.fetchmany()
            return self._cursor.fetchmany(size)
        return self._db_provider.execute_db_function(fetch_wrapper)

    def fetchall(self) -> Future[List[sqlite3.Row]]:
        def fetch_wrapper(_) -> List[sqlite3.Row]:
            return self._cursor.fetchall()
        return self._db_provider.execute_db_function(fetch_wrapper)

class SqlTableWrapper(contextlib.AbstractAsyncContextManager):
    def __init__(
        self,
        database: MoonrakerDatabase,
        table_def: SqlTableDefinition
    ) -> None:
        self._database = database
        self._table_def = table_def
        self._db_provider = database.db_provider

    @property
    def version(self) -> int:
        return self._table_def.version

    async def __aenter__(self) -> SqlTableWrapper:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_value is not None:
            await self.rollback()
        else:
            await self.commit()

    def get_provider_wrapper(self) -> DBProviderWrapper:
        return self._database.get_provider_wrapper()

    def queue_callback(
        self, callback: Callable[[sqlite3.Connection], Any]
    ) -> Future[Any]:
        return self._db_provider.execute_db_function(callback)

    def execute(
        self, sql: str, params: SqlParams = []
    ) -> Future[SqliteCursorProxy]:
        return self._db_provider.execute_db_function(
            self._db_provider.sql_execute, sql, params
        )

    def executemany(
        self, sql: str, params: Sequence[SqlParams] = []
    ) -> Future[SqliteCursorProxy]:
        return self._db_provider.execute_db_function(
            self._db_provider.sql_executemany, sql, params
        )

    def executescript(self, sql: str) -> Future[SqliteCursorProxy]:
        return self._db_provider.execute_db_function(
            self._db_provider.sql_executescript, sql
        )

    def commit(self) -> Future[None]:
        return self._db_provider.execute_db_function(
            self._db_provider.sql_commit
        )

    def rollback(self) -> Future[None]:
        return self._db_provider.execute_db_function(
            self._db_provider.sql_rollback
        )


class NamespaceWrapper:
    def __init__(
        self,
        namespace: str,
        database: MoonrakerDatabase,
        parse_keys: bool = False
    ) -> None:
        self.namespace = namespace
        self.db = database
        self.eventloop = database.eventloop
        self.server = database.server
        # If parse keys is true, keys of a string type
        # will be passed straight to the DB methods.
        self._parse_keys = parse_keys

    @property
    def parse_keys(self) -> bool:
        return self._parse_keys

    @parse_keys.setter
    def parse_keys(self, val: bool) -> None:
        self._parse_keys = val

    def get_provider_wrapper(self) -> DBProviderWrapper:
        return self.db.get_provider_wrapper()

    def insert(
        self, key: Union[List[str], str], value: DBType
    ) -> Future[None]:
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.insert_item(self.namespace, key, value)

    def update_child(
        self, key: Union[List[str], str], value: DBType
    ) -> Future[None]:
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.update_item(self.namespace, key, value)

    def update(self, value: Dict[str, DBRecord]) -> Future[None]:
        return self.db.update_namespace(self.namespace, value)

    def sync(self, value: Dict[str, DBRecord]) -> Future[None]:
        return self.db.sync_namespace(self.namespace, value)

    def get(self, key: Union[List[str], str], default: Any = None) -> Future[Any]:
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.get_item(self.namespace, key, default)

    def delete(self, key: Union[List[str], str]) -> Future[Any]:
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.delete_item(self.namespace, key)

    def insert_batch(self, records: Dict[str, Any]) -> Future[None]:
        return self.db.insert_batch(self.namespace, records)

    def move_batch(self,
                   source_keys: List[str],
                   dest_keys: List[str]
                   ) -> Future[None]:
        return self.db.move_batch(self.namespace, source_keys, dest_keys)

    def delete_batch(self, keys: List[str]) -> Future[Dict[str, Any]]:
        return self.db.delete_batch(self.namespace, keys)

    def get_batch(self, keys: List[str]) -> Future[Dict[str, Any]]:
        return self.db.get_batch(self.namespace, keys)

    def length(self) -> Future[int]:
        return self.db.ns_length(self.namespace)

    def as_dict(self) -> Dict[str, Any]:
        self._check_sync_method("as_dict")
        return self.db.get_item(self.namespace).result()

    def __getitem__(self, key: Union[List[str], str]) -> Future[Any]:
        return self.get(key, default=Sentinel.MISSING)

    def __setitem__(self, key: Union[List[str], str], value: DBType) -> None:
        self.insert(key, value)

    def __delitem__(self, key: Union[List[str], str]):
        self.delete(key)

    def __contains__(self, key: Union[List[str], str]) -> bool:
        self._check_sync_method("__contains__")
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.ns_contains(self.namespace, key).result()

    def contains(self, key: Union[List[str], str]) -> Future[bool]:
        if isinstance(key, str) and not self._parse_keys:
            key = [key]
        return self.db.ns_contains(self.namespace, key)

    def keys(self) -> Future[List[str]]:
        return self.db.ns_keys(self.namespace)

    def values(self) -> Future[List[Any]]:
        return self.db.ns_values(self.namespace)

    def items(self) -> Future[List[Tuple[str, Any]]]:
        return self.db.ns_items(self.namespace)

    def pop(
        self, key: Union[List[str], str], default: Any = Sentinel.MISSING
    ) -> Union[Future[Any], Task[Any]]:
        if not self.server.is_running():
            try:
                val = self.delete(key).result()
            except Exception:
                if default is Sentinel.MISSING:
                    raise
                val = default
            fut = self.eventloop.create_future()
            fut.set_result(val)
            return fut

        async def _do_pop() -> Any:
            try:
                val = await self.delete(key)
            except Exception:
                if default is Sentinel.MISSING:
                    raise
                val = default
            return val
        return self.eventloop.create_task(_do_pop())

    def clear(self) -> Future[None]:
        return self.db.clear_namespace(self.namespace)

    def _check_sync_method(self, func_name: str) -> None:
        if self.db.db_provider.is_alive():
            raise self.server.error(
                f"Cannot call method {func_name} while "
                "the eventloop is running"
            )

def load_component(config: ConfigHelper) -> MoonrakerDatabase:
    return MoonrakerDatabase(config)



# Moonraker extension management
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import asyncio
import pathlib
import logging
from ..common import BaseRemoteConnection, RequestType, TransportType
from ..utils import get_unix_peer_credentials

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Union,
)

if TYPE_CHECKING:
    from ..server import Server
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .klippy_connection import KlippyConnection as Klippy

UNIX_BUFFER_LIMIT = 20 * 1024 * 1024

class ExtensionManager:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.agents: Dict[str, BaseRemoteConnection] = {}
        self.agent_methods: Dict[int, List[str]] = {}
        self.uds_server: Optional[asyncio.AbstractServer] = None
        self.server.register_endpoint(
            "/connection/register_remote_method", RequestType.POST,
            self._register_agent_method,
            transports=TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/connection/send_event", RequestType.POST, self._handle_agent_event,
            transports=TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/server/extensions/list", RequestType.GET, self._handle_list_extensions
        )
        self.server.register_endpoint(
            "/server/extensions/request", RequestType.POST, self._handle_call_agent
        )

    def register_agent(self, connection: BaseRemoteConnection) -> None:
        data = connection.client_data
        name = data["name"]
        client_type = data["type"]
        if client_type != "agent":
            raise self.server.error(
                f"Cannot register client type '{client_type}' as an agent"
            )
        if name in self.agents:
            raise self.server.error(
                f"Agent '{name}' already registered and connected'"
            )
        self.agents[name] = connection
        data = connection.client_data
        evt: Dict[str, Any] = {
            "agent": name, "event": "connected", "data": data
        }
        connection.send_notification("agent_event", [evt])

    def remove_agent(self, connection: BaseRemoteConnection) -> None:
        name = connection.client_data["name"]
        if name in self.agents:
            klippy: Klippy = self.server.lookup_component("klippy_connection")
            registered_methods = self.agent_methods.pop(connection.uid, [])
            for method in registered_methods:
                klippy.unregister_method(method)
            del self.agents[name]
            evt: Dict[str, Any] = {"agent": name, "event": "disconnected"}
            connection.send_notification("agent_event", [evt])

    async def _handle_agent_event(self, web_request: WebRequest) -> str:
        conn = web_request.get_client_connection()
        if conn is None:
            raise self.server.error("No connection detected")
        if conn.client_data["type"] != "agent":
            raise self.server.error(
                "Only connections of the 'agent' type can send events"
            )
        name = conn.client_data["name"]
        evt_name = web_request.get_str("event")
        if evt_name in ["connected", "disconnected"]:
            raise self.server.error(f"Event '{evt_name}' is reserved")
        data: Optional[Union[List, Dict[str, Any]]]
        data = web_request.get("data", None)
        evt: Dict[str, Any] = {"agent": name, "event": evt_name}
        if data is not None:
            evt["data"] = data
        conn.send_notification("agent_event", [evt])
        return "ok"

    async def _register_agent_method(self, web_request: WebRequest) -> str:
        conn = web_request.get_client_connection()
        if conn is None:
            raise self.server.error("No connection detected")
        method_name = web_request.get_str("method_name")
        klippy: Klippy = self.server.lookup_component("klippy_connection")
        klippy.register_method_from_agent(conn, method_name)
        self.agent_methods.setdefault(conn.uid, []).append(method_name)
        return "ok"

    async def _handle_list_extensions(
        self, web_request: WebRequest
    ) -> Dict[str, List[Dict[str, Any]]]:
        agents: List[Dict[str, Any]]
        agents = [agt.client_data for agt in self.agents.values()]
        return {"agents": agents}

    async def _handle_call_agent(self, web_request: WebRequest) -> Any:
        agent = web_request.get_str("agent")
        method: str = web_request.get_str("method")
        args: Optional[Union[List, Dict[str, Any]]]
        args = web_request.get("arguments", None)
        if args is not None and not isinstance(args, (list, dict)):
            raise self.server.error(
                "The 'arguments' field must contain an object or a list"
            )
        if agent not in self.agents:
            raise self.server.error(f"Agent {agent} not connected")
        conn = self.agents[agent]
        return await conn.call_method_with_response(method, args)

    async def start_unix_server(self) -> None:
        sockfile: str = self.server.get_app_args()["unix_socket_path"]
        sock_path = pathlib.Path(sockfile).expanduser().resolve()
        logging.info(f"Creating Unix Domain Socket at '{sock_path}'")
        try:
            self.uds_server = await asyncio.start_unix_server(
                self.on_unix_socket_connected, sock_path, limit=UNIX_BUFFER_LIMIT
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.exception(f"Failed to create Unix Domain Socket: {sock_path}")
            self.uds_server = None

    def on_unix_socket_connected(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        peercred = get_unix_peer_credentials(writer, "Unix Client Connection")
        UnixSocketClient(self.server, reader, writer, peercred)

    async def close(self) -> None:
        if self.uds_server is not None:
            self.uds_server.close()
            await self.uds_server.wait_closed()
            self.uds_server = None

class UnixSocketClient(BaseRemoteConnection):
    def __init__(
        self,
        server: Server,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        peercred: Dict[str, int]
    ) -> None:
        self.on_create(server)
        self.writer = writer
        self._peer_cred = peercred
        self._connected_time = self.eventloop.get_loop_time()
        pid = self._peer_cred.get("process_id")
        uid = self._peer_cred.get("user_id")
        gid = self._peer_cred.get("group_id")
        self.wsm.add_client(self)
        logging.info(
            f"Unix Socket Opened - Client ID: {self.uid}, "
            f"Process ID: {pid}, User ID: {uid},  Group ID: {gid}"
        )
        self.eventloop.register_callback(self._read_messages, reader)

    async def _read_messages(self, reader: asyncio.StreamReader) -> None:
        errors_remaining: int = 10
        while not reader.at_eof():
            try:
                data = await reader.readuntil(b'\x03')
                decoded = data[:-1].decode(encoding="utf-8")
            except (ConnectionError, asyncio.IncompleteReadError):
                break
            except asyncio.CancelledError:
                logging.exception("Unix Client Stream Read Cancelled")
                raise
            except Exception:
                logging.exception("Unix Client Stream Read Error")
                errors_remaining -= 1
                if not errors_remaining or self.is_closed:
                    break
                continue
            errors_remaining = 10
            self.eventloop.register_callback(self._process_message, decoded)
        logging.debug("Unix Socket Disconnection From _read_messages()")
        await self._on_close(reason="Read Exit")

    async def write_to_socket(self, message: Union[bytes, str]) -> None:
        if isinstance(message, str):
            data = message.encode() + b"\x03"
        else:
            data = message + b"\x03"
        try:
            self.writer.write(data)
            await self.writer.drain()
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.debug("Unix Socket Disconnection From write_to_socket()")
            await self._on_close(reason="Write Exception")

    async def _on_close(
        self,
        code: Optional[int] = None,
        reason: Optional[str] = None
    ) -> None:
        if self.is_closed:
            return
        self.is_closed = True
        kconn: Klippy = self.server.lookup_component("klippy_connection")
        kconn.remove_subscription(self)
        if not self.writer.is_closing():
            self.writer.close()
            try:
                await self.writer.wait_closed()
            except Exception:
                pass
        self.message_buf = []
        for resp in self.pending_responses.values():
            resp.set_exception(
                self.server.error("Client Socket Disconnected", 500)
            )
        self.pending_responses = {}
        logging.info(
            f"Unix Socket Closed: ID: {self.uid}, "
            f"Close Code: {code}, "
            f"Close Reason: {reason}"
        )
        if self._client_data["type"] == "agent":
            extensions: ExtensionManager
            extensions = self.server.lookup_component("extensions")
            extensions.remove_agent(self)
        self.wsm.remove_client(self)

    def close_socket(self, code: int, reason: str) -> None:
        if not self.is_closed:
            self.eventloop.register_callback(self._on_close, code, reason)


def load_component(config: ConfigHelper) -> ExtensionManager:
    return ExtensionManager(config)
